"""Policies for the PursuitEvasion-v0 environment."""
import os.path as osp

from posggym.agents.grid_world.pursuit_evasion.shortest_path import PEShortestPathPolicy
from posggym.agents.registration import PolicySpec
from posggym.agents.torch_policy import PPOPolicy
from posggym.agents.utils import processors
from posggym.config import AGENT_MODEL_DIR


ENV_ID = "PursuitEvasion-v0"
agent_model_dir = osp.join(AGENT_MODEL_DIR, "grid_world", "pursuit_evasion")
policy_specs = {}

_shortest_path_spec = PolicySpec(
    policy_name="shortestpath",
    entry_point=PEShortestPathPolicy,
    version=0,
    env_id="PursuitEvasion-v0",
    env_args=None,
    valid_agent_ids=None,
    nondeterministic=False,
)
policy_specs[_shortest_path_spec.id] = _shortest_path_spec

# PursuitEvasion 16x16
# Evader (agent=0)
for policy_file_name in [
    "klr_k0_seed0_i0.pkl",
    "klr_k0_seed1_i0.pkl",
    "klr_k0_seed2_i0.pkl",
    "klr_k0_seed3_i0.pkl",
    "klr_k0_seed4_i0.pkl",
    "klr_k1_seed0_i0.pkl",
    "klr_k1_seed1_i0.pkl",
    "klr_k1_seed2_i0.pkl",
    "klr_k1_seed3_i0.pkl",
    "klr_k1_seed4_i0.pkl",
    "klr_k2_seed0_i0.pkl",
    "klr_k2_seed1_i0.pkl",
    "klr_k2_seed2_i0.pkl",
    "klr_k2_seed3_i0.pkl",
    "klr_k2_seed4_i0.pkl",
    "klr_k3_seed0_i0.pkl",
    "klr_k3_seed1_i0.pkl",
    "klr_k3_seed2_i0.pkl",
    "klr_k3_seed3_i0.pkl",
    "klr_k3_seed4_i0.pkl",
    "klr_k4_seed0_i0.pkl",
    "klr_k4_seed1_i0.pkl",
    "klr_k4_seed2_i0.pkl",
    "klr_k4_seed3_i0.pkl",
    "klr_k4_seed4_i0.pkl",
    "klrbr_k4_seed0_i0.pkl",
    "klrbr_k4_seed1_i0.pkl",
    "klrbr_k4_seed2_i0.pkl",
    "klrbr_k4_seed3_i0.pkl",
    "klrbr_k4_seed4_i0.pkl",
    "sp_seed0_i0.pkl",
    "sp_seed1_i0.pkl",
    "sp_seed2_i0.pkl",
    "sp_seed3_i0.pkl",
    "sp_seed4_i0.pkl",
]:
    spec = PPOPolicy.get_spec_from_path(
        env_id=ENV_ID,
        env_args={"grid": "16x16"},
        policy_file_path=osp.join(agent_model_dir, "grid=16x16", policy_file_name),
        version=0,
        valid_agent_ids=["0"],
        # policy is deterministic given random seed
        nondeterministic=False,
        # actions sampled, rather always taking most probable action
        deterministic=False,
        obs_processor_cls=processors.FlattenProcessor,
        obs_processor_config=None,
        action_processor_cls=None,
        action_processor_config=None,
    )
    policy_specs[spec.id] = spec


# Pursuer (agent=1)
for policy_file_name in [
    "klr_k0_seed0_i1.pkl",
    "klr_k0_seed1_i1.pkl",
    "klr_k0_seed2_i1.pkl",
    "klr_k0_seed3_i1.pkl",
    "klr_k0_seed4_i1.pkl",
    "klr_k1_seed0_i1.pkl",
    "klr_k1_seed1_i1.pkl",
    "klr_k1_seed2_i1.pkl",
    "klr_k1_seed3_i1.pkl",
    "klr_k1_seed4_i1.pkl",
    "klr_k2_seed0_i1.pkl",
    "klr_k2_seed1_i1.pkl",
    "klr_k2_seed2_i1.pkl",
    "klr_k2_seed3_i1.pkl",
    "klr_k2_seed4_i1.pkl",
    "klr_k3_seed0_i1.pkl",
    "klr_k3_seed1_i1.pkl",
    "klr_k3_seed2_i1.pkl",
    "klr_k3_seed3_i1.pkl",
    "klr_k3_seed4_i1.pkl",
    "klr_k4_seed0_i1.pkl",
    "klr_k4_seed1_i1.pkl",
    "klr_k4_seed2_i1.pkl",
    "klr_k4_seed3_i1.pkl",
    "klr_k4_seed4_i1.pkl",
    "klrbr_k4_seed0_i1.pkl",
    "klrbr_k4_seed1_i1.pkl",
    "klrbr_k4_seed2_i1.pkl",
    "klrbr_k4_seed3_i1.pkl",
    "klrbr_k4_seed4_i1.pkl",
    "sp_seed0_i1.pkl",
    "sp_seed1_i1.pkl",
    "sp_seed2_i1.pkl",
    "sp_seed3_i1.pkl",
    "sp_seed4_i1.pkl",
]:
    spec = PPOPolicy.get_spec_from_path(
        env_id=ENV_ID,
        env_args={"grid": "16x16"},
        policy_file_path=osp.join(agent_model_dir, "grid=16x16", policy_file_name),
        version=0,
        valid_agent_ids=["1"],
        # policy is deterministic given random seed
        nondeterministic=False,
        # actions sampled, rather always taking most probable action
        deterministic=False,
        obs_processor_cls=processors.FlattenProcessor,
        obs_processor_config=None,
        action_processor_cls=None,
        action_processor_config=None,
    )
    policy_specs[spec.id] = spec

# PursuitEvasion 8x8
# Evader (agent=0)
for policy_file_name in [
    "klr_k0_seed0_i0.pkl",
    "klr_k0_seed1_i0.pkl",
    "klr_k0_seed2_i0.pkl",
    "klr_k0_seed3_i0.pkl",
    "klr_k0_seed4_i0.pkl",
    "klr_k1_seed0_i0.pkl",
    "klr_k1_seed1_i0.pkl",
    "klr_k1_seed2_i0.pkl",
    "klr_k1_seed3_i0.pkl",
    "klr_k1_seed4_i0.pkl",
    "klr_k2_seed0_i0.pkl",
    "klr_k2_seed1_i0.pkl",
    "klr_k2_seed2_i0.pkl",
    "klr_k2_seed3_i0.pkl",
    "klr_k2_seed4_i0.pkl",
    "klr_k3_seed0_i0.pkl",
    "klr_k3_seed1_i0.pkl",
    "klr_k3_seed2_i0.pkl",
    "klr_k3_seed3_i0.pkl",
    "klr_k3_seed4_i0.pkl",
    "klr_k4_seed0_i0.pkl",
    "klr_k4_seed1_i0.pkl",
    "klr_k4_seed2_i0.pkl",
    "klr_k4_seed3_i0.pkl",
    "klr_k4_seed4_i0.pkl",
    "klrbr_k4_seed0_i0.pkl",
    "klrbr_k4_seed1_i0.pkl",
    "klrbr_k4_seed2_i0.pkl",
    "klrbr_k4_seed3_i0.pkl",
    "klrbr_k4_seed4_i0.pkl",
    "sp_seed0_i0.pkl",
    "sp_seed1_i0.pkl",
    "sp_seed2_i0.pkl",
    "sp_seed3_i0.pkl",
    "sp_seed4_i0.pkl",
]:
    spec = PPOPolicy.get_spec_from_path(
        env_id=ENV_ID,
        env_args={"grid": "8x8"},
        policy_file_path=osp.join(agent_model_dir, "grid=8x8", policy_file_name),
        version=0,
        valid_agent_ids=["0"],
        # policy is deterministic given random seed
        nondeterministic=False,
        # actions sampled, rather always taking most probable action
        deterministic=False,
        obs_processor_cls=processors.FlattenProcessor,
        obs_processor_config=None,
        action_processor_cls=None,
        action_processor_config=None,
    )
    policy_specs[spec.id] = spec


# Pursuer (agent=1)
for policy_file_name in [
    "klr_k0_seed0_i1.pkl",
    "klr_k0_seed1_i1.pkl",
    "klr_k0_seed2_i1.pkl",
    "klr_k0_seed3_i1.pkl",
    "klr_k0_seed4_i1.pkl",
    "klr_k1_seed0_i1.pkl",
    "klr_k1_seed1_i1.pkl",
    "klr_k1_seed2_i1.pkl",
    "klr_k1_seed3_i1.pkl",
    "klr_k1_seed4_i1.pkl",
    "klr_k2_seed0_i1.pkl",
    "klr_k2_seed1_i1.pkl",
    "klr_k2_seed2_i1.pkl",
    "klr_k2_seed3_i1.pkl",
    "klr_k2_seed4_i1.pkl",
    "klr_k3_seed0_i1.pkl",
    "klr_k3_seed1_i1.pkl",
    "klr_k3_seed2_i1.pkl",
    "klr_k3_seed3_i1.pkl",
    "klr_k3_seed4_i1.pkl",
    "klr_k4_seed0_i1.pkl",
    "klr_k4_seed1_i1.pkl",
    "klr_k4_seed2_i1.pkl",
    "klr_k4_seed3_i1.pkl",
    "klr_k4_seed4_i1.pkl",
    "klrbr_k4_seed0_i1.pkl",
    "klrbr_k4_seed1_i1.pkl",
    "klrbr_k4_seed2_i1.pkl",
    "klrbr_k4_seed3_i1.pkl",
    "klrbr_k4_seed4_i1.pkl",
    "sp_seed0_i1.pkl",
    "sp_seed1_i1.pkl",
    "sp_seed2_i1.pkl",
    "sp_seed3_i1.pkl",
    "sp_seed4_i1.pkl",
]:
    spec = PPOPolicy.get_spec_from_path(
        env_id=ENV_ID,
        env_args={"grid": "8x8"},
        policy_file_path=osp.join(agent_model_dir, "grid=8x8", policy_file_name),
        version=0,
        valid_agent_ids=["1"],
        # policy is deterministic given random seed
        nondeterministic=False,
        # actions sampled, rather always taking most probable action
        deterministic=False,
        obs_processor_cls=processors.FlattenProcessor,
        obs_processor_config=None,
        action_processor_cls=None,
        action_processor_config=None,
    )
    policy_specs[spec.id] = spec



"""Wrapper for converting a posggym environment into pettingzoo environment."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple

import posggym

try:
    from pettingzoo.utils.env import ActionDict, ObsDict, ParallelEnv
except ImportError as e:
    raise posggym.error.DependencyNotInstalled(
        "pettingzoo is not installed, run `pip install pettingzoo` or visit "
        "'https://github.com/Farama-Foundation/PettingZoo#installation' for details on "
        "installing pettingzoo."
    ) from e


if TYPE_CHECKING:
    import numpy as np
    from gymnasium import spaces


class PettingZoo(ParallelEnv):
    """Converts POSGGym environment into a PettingZoo environment.

    Converts into a ``pettingzoo.ParallelEnv`` environment.

    This involves:

    - treating agent IDs as strings (instead of ints in some environments)
    - handling case where an individual agent is done in environment before the episode
      is over

    References
    ----------

    - parallel env docs: https://pettingzoo.farama.org/api/parallel/
    - parallel env code:
      https://github.com/Farama-Foundation/PettingZoo/blob/master/pettingzoo/utils/env.py

    """

    def __init__(self, env: posggym.Env):
        self.env = env
        self._done_agents: Set[str] = set()

    @property
    def metadata(self) -> Dict[str, Any]:
        return self.env.metadata

    @property
    def agents(self) -> List[str]:
        return [i for i in self.env.agents if i not in self._done_agents]

    @property
    def possible_agents(self) -> List[str]:
        return list(self.env.possible_agents)

    @property
    def action_spaces(self) -> Dict[str, spaces.Space]:
        return self.env.action_spaces

    @property
    def observation_spaces(self) -> Dict[str, spaces.Space]:
        return self.env.observation_spaces

    @property
    def render_mode(self) -> str | None:
        return self.env.render_mode

    def reset(
        self,
        seed: Optional[int] = None,
        return_info: bool = False,
        options: Optional[dict] = None,
    ) -> ObsDict:
        obs, info = self.env.reset(seed=seed, options=options)
        self._done_agents = set()

        if return_info:
            return obs, info
        return obs

    def step(
        self, actions: ActionDict
    ) -> Tuple[
        ObsDict, Dict[str, float], Dict[str, bool], Dict[str, bool], Dict[str, dict]
    ]:
        obs, rewards, terminated, truncated, all_done, info = self.env.step(actions)

        # remove obs for already done agents, since pettingzoo expects no output for
        # agents that were terminated or truncated in a previous step
        for d in (obs, rewards, terminated, truncated, info):
            for i in self._done_agents:
                d.pop(i, None)

        # add any newly done agents to set of done agents so they are handled
        # appropriately next step
        for i, done in terminated.items():
            if done:
                self._done_agents.add(i)
        for i, done in truncated.items():
            if done:
                self._done_agents.add(i)

        return obs, rewards, terminated, truncated, info

    def render(self) -> None | np.ndarray | str | List:
        output = self.env.render()
        if isinstance(output, dict):
            return output.get("env", None)
        return output

    def close(self):
        self.env.close()

    def state(self):
        return self.env.state

    def observation_space(self, agent: str) -> spaces.Space:
        return self.observation_spaces[agent]

    def action_space(self, agent: str) -> spaces.Space:
        return self.action_spaces[agent]



"""Sub-package containing code for running evaluation experiments on policies."""
from posggym.agents.evaluation.exp import ExpParams, get_exp_parser, run_experiments
from posggym.agents.evaluation.pairwise import run_pairwise_experiments
from posggym.agents.evaluation.render import EpisodeRenderer, Renderer
from posggym.agents.evaluation.runner import run_episode
from posggym.agents.evaluation.stats import get_default_trackers



"""Policies for the PursuitEvasionContinuous-v0 environment."""
import math
import os.path as osp

from posggym.agents.continuous.pursuit_evasion_continuous import shortest_path
from posggym.agents.registration import PolicySpec
from posggym.agents.torch_policy import PPOPolicy
from posggym.agents.utils import processors
from posggym.config import AGENT_MODEL_DIR

ENV_ID = "PursuitEvasionContinuous-v0"
agent_model_dir = osp.join(AGENT_MODEL_DIR, "continuous", "pursuit_evasion_continuous")
policy_specs = {}

_sp_spec = PolicySpec(
    policy_name="shortest_path",
    entry_point=shortest_path.PECShortestPathPolicy,
    version=0,
    env_id=ENV_ID,
    env_args=None,
    valid_agent_ids=None,
    nondeterministic=False,
)
policy_specs[_sp_spec.id] = _sp_spec

# 8x8 self-play trained policies with default env args
for policy_file_name in [
    "sp_seed0_i0.pkl",
    "sp_seed0_i1.pkl",
    "sp_seed1_i0.pkl",
    "sp_seed1_i1.pkl",
    "sp_seed2_i0.pkl",
    "sp_seed2_i1.pkl",
    "sp_seed3_i0.pkl",
    "sp_seed3_i1.pkl",
    "sp_seed4_i0.pkl",
    "sp_seed4_i1.pkl",
]:
    spec = PPOPolicy.get_spec_from_path(
        policy_file_path=osp.join(
            agent_model_dir,
            "world=8x8",
            policy_file_name,
        ),
        env_id=ENV_ID,
        env_args={
            "world": "8x8",
            "fov": math.pi / 3,
            "max_obs_distance": 8.0 / 3,
            "n_sensors": 16,
        },
        env_args_id="world=8x8",
        version=0,
        valid_agent_ids=["0" if "_i0" in policy_file_name else "1"],
        # policy is deterministic given random seed
        nondeterministic=False,
        # actions sampled, rather always taking most probable action
        deterministic=False,
        obs_processor_cls=processors.RescaleProcessor,
        obs_processor_config={"min_val": -1, "max_val": 1},
        action_processor_cls=processors.RescaleProcessor,
        action_processor_config={"min_val": -1, "max_val": 1},
    )
    policy_specs[spec.id] = spec

# 16x16 self-play trained policies with default env args
for policy_file_name in [
    "sp_seed0_i0.pkl",
    "sp_seed0_i1.pkl",
    "sp_seed1_i0.pkl",
    "sp_seed1_i1.pkl",
    "sp_seed2_i0.pkl",
    "sp_seed2_i1.pkl",
    "sp_seed3_i0.pkl",
    "sp_seed3_i1.pkl",
    "sp_seed4_i0.pkl",
    "sp_seed4_i1.pkl",
]:
    spec = PPOPolicy.get_spec_from_path(
        policy_file_path=osp.join(
            agent_model_dir,
            "world=16x16",
            policy_file_name,
        ),
        env_id=ENV_ID,
        env_args={
            "world": "16x16",
            "fov": math.pi / 3,
            "max_obs_distance": 16.0 / 3,
            "n_sensors": 16,
        },
        env_args_id="world=16x16",
        version=0,
        valid_agent_ids=["0" if "_i0" in policy_file_name else "1"],
        # policy is deterministic given random seed
        nondeterministic=False,
        # actions sampled, rather always taking most probable action
        deterministic=False,
        obs_processor_cls=processors.RescaleProcessor,
        obs_processor_config={"min_val": -1, "max_val": 1},
        action_processor_cls=processors.RescaleProcessor,
        action_processor_config={"min_val": -1, "max_val": 1},
    )
    policy_specs[spec.id] = spec



"""Core functionality for continuous environments."""
from __future__ import annotations

import math
import warnings
from abc import ABC, abstractmethod
from enum import Enum
from itertools import product
from queue import PriorityQueue
from typing import Dict, Iterable, List, NamedTuple, Set, Tuple, Union

import numpy as np
from gymnasium import spaces

from posggym.error import DependencyNotInstalled

try:
    import pymunk
    from pymunk import Vec2d
except ImportError as e:
    raise DependencyNotInstalled(
        "pymunk is not installed, run `pip install posggym[continuous]`"
    ) from e


# (x, y) coord = (col, row) coord
Coord = Tuple[int, int]
FloatCoord = Tuple[float, float]
# (x, y, yaw) in continuous world
Position = Tuple[float, float, float]
Location = Union[Coord, FloatCoord, Position, np.ndarray]
# Position, radius
CircleEntity = Tuple[Position, float]
# start (x, y), end (x, y)
Line = Tuple[FloatCoord, FloatCoord]
IntLine = Tuple[Tuple[int, int], Tuple[int, int]]

# (Main, Alternative) agent colors (from pygame.colordict.THECOLORS)
AGENT_COLORS = [
    ((255, 0, 0, 255), (205, 0, 0, 255)),  # red, red3
    ((0, 0, 255, 255), (0, 0, 205, 255)),  # blue, blue3
    ((0, 255, 0, 255), (0, 205, 0, 255)),  # green, green3
    ((160, 32, 240, 255), (125, 38, 205, 255)),  # purple, purple3
    ((255, 255, 0, 255), (205, 205, 0, 255)),  # yellow, yellow3
    ((0, 255, 255, 255), (0, 205, 205, 255)),  # cyan, cyan3
    ((255, 105, 180, 255), (205, 96, 144, 255)),  # hotpink, hotpink3
    ((255, 165, 0, 255), (205, 133, 0, 255)),  # orange, orange3
]


class CollisionType(Enum):
    """Type of collision in world."""

    NONE = 0
    AGENT = 1
    BLOCK = 2
    BORDER = 3
    INTERIOR_WALL = 4


class PMBodyState(NamedTuple):
    """State of a Pymunk Body."""

    x: float
    y: float
    angle: float  # in radians
    vx: float
    vy: float
    vangle: float  # in radians/s

    @staticmethod
    def num_features() -> int:
        """Get the number of features of a pymunk body state."""
        return 6

    @staticmethod
    def get_space(world_size: float) -> spaces.Box:
        """Get the space for a pymunk body's state."""
        # x, y, angle, vx, vy, vangle
        # shape = (1, 6)
        size, angle = world_size, 2 * math.pi
        low = np.array([-1, -1, -angle, -1, -1, -angle], dtype=np.float32)
        high = np.array(
            [size, size, angle, 1.0, 1.0, angle],
            dtype=np.float32,
        )
        return spaces.Box(low=low, high=high)


# This function needs to be in global scope or we get pickle errors
def ignore_collisions(arbiter, space, data):
    """Pymunk collision handler which ignores collisions."""
    return False


def clip_actions(
    actions: Dict[str, np.ndarray], action_spaces: Dict[str, spaces.Space]
) -> Dict[str, np.ndarray]:
    """Clip continuous actions so they are within the agents action space dims."""
    clipped_actions = {}
    for i, a in actions.items():
        a_space = action_spaces[i]
        assert isinstance(a_space, spaces.Box)
        clipped_actions[i] = np.clip(a, a_space.low, a_space.high)
    return clipped_actions


class AbstractContinuousWorld(ABC):
    """A continuous 2D world with a rectangular border."""

    WALL_COLOR = (0, 0, 0, 255)  # Black
    BLOCK_COLOR = (50, 50, 50, 255)  # Black

    def __init__(
        self,
        size: float,
        blocks: List[CircleEntity] | None = None,
        interior_walls: List[Line] | None = None,
        agent_radius: float = 0.5,
        border_thickness: float = 0.1,
        enable_agent_collisions: bool = True,
    ):
        self.size = size
        self.blocks = blocks or []
        self.interior_walls = interior_walls or []
        self.agent_radius = agent_radius
        self.border_thickness = border_thickness
        # access via blocked_coords property
        self._blocked_coords: Set[Coord] | None = None

        self.collision_id = 0
        self.enable_agent_collisions = enable_agent_collisions

        # 2D physics stuff
        self.space = pymunk.Space()
        self.space.gravity = Vec2d(0.0, 0.0)

        if not enable_agent_collisions:
            # Turn off all collisions
            self.space.add_collision_handler(0, 0).pre_solve = ignore_collisions

        self.add_border_to_space(size)
        self.add_interior_walls_to_space(self.interior_walls)

        for pos, radius in self.blocks:
            body = pymunk.Body(body_type=pymunk.Body.STATIC)
            shape = pymunk.Circle(body, radius)
            body.position = Vec2d(pos[0], pos[1])
            shape.elasticity = 0.0  # no bouncing
            shape.color = self.BLOCK_COLOR
            shape.collision_type = self.get_collision_id()
            self.space.add(body, shape)

        # moveable entities in the world
        self.entities: Dict[str, Tuple[pymunk.Body, pymunk.Circle]] = {}

    @abstractmethod
    def add_border_to_space(self, size: float):
        """Adds solid border to the world physics space."""
        pass

    @abstractmethod
    def check_border_collisions(
        self, ray_start_coords: np.ndarray, ray_end_coords: np.ndarray
    ) -> np.ndarray:
        """Check for collision between rays and world border."""
        pass

    @abstractmethod
    def clip_position(self, position: Vec2d) -> Vec2d:
        """Clip the position of an agent to be inside the border"""
        pass

    @abstractmethod
    def copy(self) -> "AbstractContinuousWorld":
        """Get a deep copy of this world."""
        pass

    def simulate(
        self,
        dt: float = 1.0 / 10,
        t: int = 10,
        normalize_angles: bool = True,
    ):
        """Simulate the world, updating all entities.

        As per pymunk.Space.step documentation, using a fixed time step `dt` is
        recommended and improves efficiency.

        Also performing multiple steps `t` with a smaller `dt` creates a more stable
        and accurate simulation.

        Arguments
        ---------
        dt: the step size
        t: the number of steps
        normalize_angles: whether to normalize angle of each entity to be in [0, 2*pi]
            at the end of the simulation.

        Reference
        ---------
        https://www.pymunk.org/en/latest/pymunk.html#pymunk.Space.step

        """
        for _ in range(t):
            self.space.step(dt)

        if normalize_angles:
            for body, _ in self.entities.values():
                body.angle = self.convert_angle_to_0_2pi_interval(body.angle)
                body.angular_velocity = self.convert_angle_to_0_2pi_interval(
                    body.angular_velocity
                )

    def add_entity(
        self,
        id: str,
        radius: float | None,
        color: Tuple[int, int, int, int] | None,
        is_static: bool = False,
    ) -> Tuple[pymunk.Body, pymunk.Circle]:
        """Add moveable entity to the world.

        Arguments
        ---------
        id: the unique ID of the entity
        radius: the radius of the entity. If none uses `self.agent_radius`
        color: optional color for the entity. This only impacts rendering of the world.

        Returns
        -------
        body: underlying physics Body of the entity
        shape: the shape of the entity

        """
        if radius is None:
            radius = self.agent_radius
        mass = 1.0
        inertia = pymunk.moment_for_circle(mass, 0.0, radius)
        body_type = pymunk.Body.STATIC if is_static else pymunk.Body.DYNAMIC
        body = pymunk.Body(mass, inertia, body_type=body_type)
        shape = pymunk.Circle(body, radius)

        shape.collision_type = self.get_collision_id()

        shape.elasticity = 0.0  # no bouncing
        shape.collision_type = self.get_collision_id()
        if color is not None:
            shape.color = color

        self.space.add(body, shape)
        self.entities[id] = (body, shape)
        return body, shape

    def remove_entity(self, id: str):
        """Remove entity from the world."""
        body, shape = self.entities[id]
        self.space.remove(body, shape)
        del self.entities[id]

    def add_interior_walls_to_space(self, walls: List[Line]):
        """Adds interior walls to the world physics space."""
        self.interior_walls_array = (
            np.array([ln[0] for ln in walls], dtype=np.float32),
            np.array([ln[1] for ln in walls], dtype=np.float32),
        )

        for w_start, w_end in walls:
            wall = pymunk.Segment(
                self.space.static_body,
                w_start,
                w_end,
                self.border_thickness,
            )
            wall.friction = 1.0
            wall.collision_type = self.get_collision_id() + 1
            wall.color = self.WALL_COLOR
            self.space.add(wall)

    def get_collision_id(self):
        return self.collision_id

    def get_entity_state(self, id: str) -> PMBodyState:
        """Get underlying state of an entity in the world."""
        body, _ = self.entities[id]
        x, y = body.position
        vx, vy = body.velocity
        return PMBodyState(x, y, body.angle, vx, vy, body.angular_velocity)

    def set_entity_state(self, id: str, state: PMBodyState | np.ndarray):
        """Set the state of an entity.

        If state is a np.ndarray, then it should be 1D with values ordered the same as
        the PMBodyState namedtuple.
        """
        body, _ = self.entities[id]
        body.position = Vec2d(state[0], state[1])
        body.angle = state[2]
        body.velocity = Vec2d(state[3], state[4])
        body.angular_velocity = state[5]

    def update_entity_state(
        self,
        id: str,
        *,
        coord: FloatCoord | List[float] | np.ndarray | Vec2d | None = None,
        angle: float | None = None,
        vel: FloatCoord | List[float] | np.ndarray | Vec2d | None = None,
        vangle: float | None = None,
    ):
        """Update the state of an entity.

        Can be used to update specific parts of the entity state, while leaving the
        other components unchanged.

        """
        body, _ = self.entities[id]
        if coord is not None:
            body.position = Vec2d(coord[0], coord[1])

        if angle is not None:
            body.angle = angle

        if vel is not None:
            body.velocity = Vec2d(vel[0], vel[1])

        if vangle is not None:
            body.angular_velocity = vangle

    def get_bounds(self) -> Tuple[FloatCoord, FloatCoord]:
        """Get  (min x, max_x), (min y, max y) bounds of the world."""
        return (0, self.size), (0, self.size)

    @property
    def blocked_coords(self) -> Set[Coord]:
        """The set of all integer coordinates that contain at least part of a block."""
        if self._blocked_coords is None:
            self._blocked_coords = set()
            # offset from coord, to middle of cell
            offset = 0.5
            # to handle any weird rounding
            max_coord = math.ceil(round(self.size, 6))
            for x, y in product(list(range(max_coord)), repeat=2):
                offset_coord = (x + offset, y + offset)

                # check if cell contains any block
                for pos, r in self.blocks:
                    if self.euclidean_dist(offset_coord, pos) < (offset + r):
                        self._blocked_coords.add((x, y))
                        break

                # check if cell contains any wall
                if len(self.interior_walls):
                    dists = self.check_circle_line_intersection(
                        np.array([[x, y]]),
                        np.array([offset]),
                        *self.interior_walls_array,
                    )
                    if np.any(np.where(dists < offset, True, False)):
                        self._blocked_coords.add((x, y))
                        break

        return self._blocked_coords

    @staticmethod
    def manhattan_dist(loc1: Location, loc2: Location) -> float:
        """Get manhattan distance between two positions in the world."""
        return abs(loc1[0] - loc2[0]) + abs(loc1[1] - loc2[1])

    @staticmethod
    def euclidean_dist(loc1: Location, loc2: Location) -> float:
        """Get Euclidean distance between two positions on the grid."""
        return math.sqrt((loc1[0] - loc2[0]) ** 2 + (loc1[1] - loc2[1]) ** 2)

    @staticmethod
    def squared_euclidean_dist(loc1: Location, loc2: Location) -> float:
        """Get Squared Euclidean distance between two positions on the grid."""
        return (loc1[0] - loc2[0]) ** 2 + (loc1[1] - loc2[1]) ** 2

    @staticmethod
    def convert_angle_to_0_2pi_interval(angle: float) -> float:
        return angle % (2 * math.pi)

    @staticmethod
    def convert_angle_to_negpi_pi_interval(angle: float) -> float:
        """Convert angle in radians to be in (-pi, pi] interval."""
        angle = angle % (2 * math.pi)
        if angle > math.pi:
            angle -= 2 * math.pi
        return angle

    @staticmethod
    def array_to_position(arr: np.ndarray) -> Position:
        """Convert from numpy array to tuple representation of a Position."""
        assert arr.shape[0] >= 3
        return (arr[0], arr[1], arr[2])

    @staticmethod
    def linear_to_xy_velocity(linear_vel: float, angle: float) -> Vec2d:
        """Convert from linear velocity to velocity along x and y axis."""
        return linear_vel * Vec2d(1, 0).rotated(angle)

    @staticmethod
    def rotate_vector(vx: float, vy: float, angle: float) -> Vec2d:
        """Rotate a 2D vector by given angle."""
        return Vec2d(vx, vy).rotated(angle)

    @staticmethod
    def clamp_norm(vx: float, vy: float, norm_max: float) -> Tuple[float, float]:
        """Clamp x, y vector to within a given max norm."""
        if vx == 0.0 and vy == 0.0:
            return vx, vy
        norm = math.sqrt(vx**2 + vy**2)
        f = min(norm, norm_max) / norm
        return f * vx, f * vy

    @staticmethod
    def convert_into_interval(
        x: float | np.ndarray,
        x_min: float | np.ndarray,
        x_max: float | np.ndarray,
        new_min: float | np.ndarray,
        new_max: float | np.ndarray,
        clip: bool = False,
    ) -> float | np.ndarray:
        """Convert variable from [x_min, x_max] to [new_min, new_max] interval."""
        new_x = (x - x_min) / (x_max - x_min) * (new_max - new_min) + new_min
        if clip:
            return np.clip(new_x, new_min, new_max)
        return new_x

    def agents_collide(self, loc1: Location, loc2: Location) -> bool:
        """Get whether two agents have collided or not.

        Assumes agents are circles with radius `self.agent_radius`.
        """
        dist = self.squared_euclidean_dist(loc1, loc2)
        return 0 <= dist <= (self.agent_radius + self.agent_radius) ** 2

    def check_circle_line_intersection(
        self,
        circle_coords: np.ndarray,
        circle_radii: np.ndarray,
        lines_start_coords: np.ndarray,
        lines_end_coords: np.ndarray,
    ) -> np.ndarray:
        """Check if lines intersect circles.

        Arguments
        ---------
        circle_coords: array containing the `(x, y)` of the center of each circle.
            Should have shape `(n_circles, 2)`
        circle_radii: array containing the radius of each circle. Should have shape
            `(n_circles,)`
        lines_start_coords: array containing the `(x, y)` coords of the start of each of
            the lines. Should have shape `(n_lines, 2)`
        lines_end_coords: array containing the `(x, y)` coords of the end of of each of
            the lines. Should have shape `(n_lines, 2)`

        Returns
        -------
        distances: An array containing the euclidean distance from each lines start to
            the first point of intersection with the corresponding circle. If the line
            does not intersect the circle, its distance will be `np.nan`.
            Should have shape `(n_circles, n_lines)`

        """
        starts = lines_start_coords[None, :, :] - circle_coords[:, None, :]
        ends = lines_end_coords[None, :, :] - circle_coords[:, None, :]
        v = ends - starts

        a = (v**2).sum(axis=2)
        b = 2 * (starts * v).sum(axis=2)
        c = (starts**2).sum(axis=2) - circle_radii[:, None] ** 2
        disc = b**2 - 4 * a * c

        with np.errstate(invalid="ignore"):
            # this will output np.nan for any negative discriminants, which is what we
            # want, but will throw a runtime warning which we want to ignore
            sqrtdisc = np.sqrt(disc)

        t1 = (-b - sqrtdisc) / (2 * a)
        t2 = (-b + sqrtdisc) / (2 * a)
        t1 = np.where(((t1 >= 0.0) & (t1 <= 1.0)), t1, np.nan)
        t2 = np.where(((t2 >= 0.0) & (t2 <= 1.0)), t2, np.nan)
        t = np.where(t1 <= t2, t1, t2)

        t = np.expand_dims(t, axis=-1)
        return np.sqrt(((t * v) ** 2).sum(axis=2))  # type: ignore

    def check_line_line_intersection(
        self,
        l1_start_coords: np.ndarray,
        l1_end_coords: np.ndarray,
        l2_start_coords: np.ndarray,
        l2_end_coords: np.ndarray,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Check if lines intersect.

        Checks for each line in `l1` if it intersects with any line in `l2`.

        Arguments
        ---------
        l1_start_coords: array with shape `(n_lines1, 2)` containing the (x, y) coord
          for the start of each of the first set of lines.
        l1_end_coords: array with shape `(n_lines1, 2)` containing the (x, y) coord
          for the end of each of the first set of lines.
        l2_start_coords: array with shape `(n_lines2, 2)` containing the (x, y) coord
          for the start of each of the second set of lines.
        l2_end_coords: array with shape `(n_lines2, 2)` containing the (x, y) coord
          for the end of each of the second set of lines.

        Returns
        -------
        intersection_coords: array with shape `(n_lines1, n_lines2, 2)` containing the
           (x, y) coords for the point of intersection between each pair of line
           segments from `l1` and `l2`. If a pair of lines does not intersect, the
           x and y coordinate values will be `np.nan`.
        distances: array with shape `(n_lines1, n_lines2)` containing the distances
            between the intersection point (if exists) and the corresponding point in
            l1_start_cooords.  If a pair of lines does not intersect, the distance value
            will be `np.nan`.

        """
        # https://stackoverflow.com/questions/3252194/numpy-and-line-intersections?noredirect=1&lq=1
        # https://en.wikipedia.org/wiki/Line%E2%80%93line_intersection
        dl1 = l1_end_coords - l1_start_coords  # (n_lines1, 2)
        dl2 = l2_end_coords - l2_start_coords  # (n_lines2, 2)

        # diff between each l1 with each l2 along x and y directions
        # (n_lines1, :, 2) - (:, n_lines2, 2) = (n_lines1, n_lines2, 2)
        dl1l2 = l1_start_coords[:, np.newaxis, :] - l2_start_coords[np.newaxis, :, :]
        # needed for matmul (n_lines1, 2, n_lines2)
        dl1l2_T = dl1l2.transpose(0, 2, 1)
        # needed for matmul (n_lines2, 2, n_lines1)
        dl1l2_T2 = dl1l2.transpose(1, 2, 0)

        # line perpendicular to each l1 line
        dl1p = np.empty_like(dl1)  # (n_lines1, 2)
        dl1p[:, 0] = -dl1[:, 1]
        dl1p[:, 1] = dl1[:, 0]

        # line perpendicular to each l2 line
        dl2p = np.empty_like(dl2)  # (n_lines1, 2)
        dl2p[:, 0] = -dl2[:, 1]
        dl2p[:, 1] = dl2[:, 0]

        # mult (n_lines1, 2) @ (n_lines1, 2, nlines2) = (n_lines1, n_lines2)
        # each i in n_lines1 is multiplied with one of the n_lines1 matrices in dl1l2
        # l1[i] @ (l1[i] - l2[j]) for i in [0, n_lines1], j in [0, n_lines2]
        u_num = np.stack([np.matmul(dl1p[i], dl1l2_T[i]) for i in range(dl1p.shape[0])])

        # mult (n_lines2, 2) @ (n_lines2, 2, nlines1) = (n_lines2, n_lines1)
        # same as above except for l2 lines
        t_num = np.stack(
            [np.matmul(dl2p[j], dl1l2_T2[j]) for j in range(dl2p.shape[0])]
        )

        # mult (n_lines1, 2) @ (2, n_lines2) = (n_lines1, n_lines2)
        # get l1[i] dot l2[j] for i in [0, n_lines1], j in [0, n_lines2]
        # but using perpendicular lines to l1,
        denom = np.matmul(dl1p, dl2.transpose())
        # handle case where lines are parallel/colinear, leading to denom being zero
        denom = np.where(np.isclose(denom, 0.0), np.nan, denom)

        # from wiki, lines only intersect if 0 <= u <= 1
        u = u_num / denom  # (n_lines1, n_lines2)
        t = t_num.transpose() / denom  # (n_lines1, n_lines2)

        # segments intersect when 0 <= u <= 1 and 0 <= t <= 1
        u = np.where(((u >= 0) & (u <= 1) & (t >= 0) & (t <= 1)), u, np.nan)
        u = u[:, :, np.newaxis]  # (n_lines1, n_lines2, 1)

        intersection_coords = u * dl2 + l2_start_coords
        intersection_distances = np.linalg.norm(
            intersection_coords - l1_start_coords[:, np.newaxis], axis=2
        )

        return intersection_coords, intersection_distances

    def check_ray_collisions(
        self,
        ray_start_coords: np.ndarray,
        ray_end_coords: np.ndarray,
        max_ray_distance: float,
        other_agents: np.ndarray | None = None,
        include_blocks: bool = True,
        check_walls: bool = True,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Check for collision along rays.

        Arguments
        ---------
        ray_start_coords: start coords of rays. Should be 2D array with shape
           `(n_rays, 2`),
        ray_end_coords: end coords of rays. Should be 2D array with shape
           `(n_rays, 2`),
        max_ray_distance: maximum ray distance
        other_agents: `(x, y)` coordinates of any other agents to check for collisions
           with. Should be a 2D array with shape `(n_agents, 2)`.
        include_blocks: whether to check for collisions with blocks in the world
        check_walls: whether to check for collisions with the world border.

        Returns
        -------
        distances: the distance each ray extends sway from the origin, up to a max of
            `max_ray_distance`. Array will have shape `(n_rays,)`.
        collision_types: the type of collision for each ray (see CollisionType), if any.
            Will have shape `(n_rays,)`.

        """
        n_rays = len(ray_start_coords)
        closest_distances = np.full(n_rays, max_ray_distance, dtype=np.float32)
        collision_types = np.full(n_rays, CollisionType.NONE.value, dtype=np.uint8)

        if other_agents is not None and len(other_agents):
            radii = np.array([self.agent_radius] * len(other_agents))
            dists = self.check_circle_line_intersection(
                other_agents, radii, ray_start_coords, ray_end_coords
            )

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                min_dists = np.nanmin(dists, axis=0)

            collision_types[min_dists < closest_distances] = CollisionType.AGENT.value
            np.fmin(closest_distances, min_dists, out=closest_distances)

        if include_blocks and len(self.blocks):
            radii = np.array([s for _, s in self.blocks])
            block_array = np.array([[pos[0], pos[1]] for pos, _ in self.blocks])

            dists = self.check_circle_line_intersection(
                block_array, radii, ray_start_coords, ray_end_coords
            )

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                min_dists = np.nanmin(dists, axis=0)

            collision_types[min_dists < closest_distances] = CollisionType.BLOCK.value
            np.fmin(closest_distances, min_dists, out=closest_distances)

        if check_walls:
            # shape = (n_lines, walls, 2)
            wall_intersect_coords = self.check_border_collisions(
                ray_start_coords, ray_end_coords
            )

            # Need to get coords of intersected walls, each ray can intersect a max of
            # of 1 wall, so we just find the minimum non nan coords
            # shape = (n_lines, 2)
            with warnings.catch_warnings():
                # if no wall intersected, we take min of all NaN which throws a warning
                # but this is acceptable behevaiour, so we suppress the warning
                warnings.simplefilter("ignore")
                wall_intersect_coords = np.nanmin(wall_intersect_coords, axis=1)

            dists = np.sqrt(
                ((wall_intersect_coords - ray_start_coords) ** 2).sum(axis=1)
            )

            collision_types[dists < closest_distances] = CollisionType.BORDER.value
            np.fmin(closest_distances, dists, out=closest_distances)

        if check_walls and len(self.interior_walls):
            _, distances = self.check_line_line_intersection(
                ray_start_coords, ray_end_coords, *self.interior_walls_array
            )

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                dists = np.nanmin(distances, axis=1)

            collision_types[
                dists < closest_distances
            ] = CollisionType.INTERIOR_WALL.value
            np.fmin(closest_distances, dists, out=closest_distances)

        return closest_distances, collision_types

    def check_collision_circular_rays(
        self,
        origin: Position,
        ray_distance: float,
        n_rays: int,
        other_agents: np.ndarray | None = None,
        include_blocks: bool = True,
        check_walls: bool = True,
        use_relative_angle: bool = True,
        angle_bounds: Tuple[float, float] = (0.0, 2 * np.pi),
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Check for collision along rays that radiate away from the origin.

        Rays are evenly spaced around the origin, with the number of rays controlled
        by the `n_rays` arguments.

        Arguments
        ---------
        origin: the origin position
        ray_distance: how far each ray extends from the origin in a straight line
        n_rays: the number of rays
        other_agents: `(x, y)` coordinates of any other agents to check for collisions
           with. Should be a 2D array with shape `(n_agents, 2)`.
        include_blocks: whether to check for collisions with blocks in the world
        check_walls: whether to check for collisions with the world border.
        use_relative_angle: whether ray angles should be relative to the origin
            position's angle. Otherwise line angle is treated as absolute (i.e.
            relative to angle of 0). This controls the ordering of the rays in the
            output.
        angle_bounds: The maximum and minimum of the FOV as a tuple. By default
            the agent will have a full FOV as a circle around them. This will
            be between 0 and 2π. This can be decreased as needed.

        Returns
        -------
        distances: the distance each ray extends sway from the origin, up to a max of
            `ray_distance`. Array will have shape `(n_rays,)`.

        """
        x, y, rel_angle = origin
        if not use_relative_angle:
            rel_angle = 0.0

        angles = np.linspace(
            angle_bounds[0], angle_bounds[1], n_rays, endpoint=False, dtype=np.float32
        )

        ray_end_xs = x + ray_distance * np.cos(angles + rel_angle)
        ray_end_ys = y + ray_distance * np.sin(angles + rel_angle)

        ray_start_coords = np.tile((x, y), (n_rays, 1))
        ray_end_coords = np.stack([ray_end_xs, ray_end_ys], axis=1)

        return self.check_ray_collisions(
            ray_start_coords,
            ray_end_coords,
            max_ray_distance=ray_distance,
            other_agents=other_agents,
            include_blocks=include_blocks,
            check_walls=check_walls,
        )

    def get_all_shortest_paths(
        self, origins: Iterable[FloatCoord | Coord | Position]
    ) -> Dict[Tuple[int, int], Dict[Tuple[int, int], int]]:
        """Get shortest path distance from every origin to all other coords."""
        src_dists = {}
        for origin in origins:
            origin_coord = self.convert_to_coord(origin)
            src_dists[origin_coord] = self.dijkstra(origin)
        return src_dists

    def convert_to_coord(self, origin: FloatCoord | Coord | Position) -> Coord:
        """Convert a position/float coord to a integer coords."""
        return (math.floor(origin[0]), math.floor(origin[1]))

    def dijkstra(self, origin: FloatCoord | Coord | Position) -> Dict[Coord, int]:
        """Get shortest path distance between origin and all other coords."""
        coord_origin = self.convert_to_coord(origin)

        dist = {coord_origin: 0}
        pq: PriorityQueue[Tuple[int, Coord]] = PriorityQueue()
        pq.put((dist[coord_origin], coord_origin))

        visited = {coord_origin}

        while not pq.empty():
            _, coord = pq.get()
            for adj_coord in self.get_cardinal_neighbours_coords(
                coord, ignore_blocks=False
            ):
                if dist[coord] + 1 < dist.get(adj_coord, float("inf")):
                    dist[adj_coord] = dist[coord] + 1

                    if adj_coord not in visited:
                        pq.put((dist[adj_coord], adj_coord))
                        visited.add(adj_coord)
        return dist

    def get_cardinal_neighbours_coords(
        self,
        coord: Coord,
        ignore_blocks: bool = False,
        include_out_of_bounds: bool = False,
    ) -> List[Coord]:
        """Get set of adjacent non-blocked coords."""
        (min_x, max_x), (min_y, max_y) = self.get_bounds()
        neighbours = []

        if coord[1] > min_y or include_out_of_bounds:
            neighbours.append((coord[0], coord[1] - 1))  # N
        if coord[0] < max_x - 1 or include_out_of_bounds:
            neighbours.append((coord[0] + 1, coord[1]))  # E
        if coord[1] < max_y or include_out_of_bounds:
            neighbours.append((coord[0], coord[1] + 1))  # S
        if coord[0] > min_x or include_out_of_bounds:
            neighbours.append((coord[0] - 1, coord[1]))  # W

        if ignore_blocks:
            return neighbours
        for i in range(len(neighbours), 0, -1):
            if neighbours[i - 1] in self.blocked_coords:
                neighbours.pop(i - 1)

        return neighbours


class SquareContinuousWorld(AbstractContinuousWorld):
    """A continuous world with a square border."""

    def copy(self) -> "SquareContinuousWorld":
        world = SquareContinuousWorld(
            size=self.size,
            blocks=self.blocks,
            interior_walls=self.interior_walls,
            agent_radius=self.agent_radius,
            border_thickness=self.border_thickness,
        )

        for id, (body, shape) in self.entities.items():
            # make copies of each entity, and ensure the copies are linked correctly
            # and added to the new world and world space
            body = body.copy()
            shape = shape.copy()
            shape.body = body

            world.space.add(body, shape)
            world.entities[id] = (body, shape)
        return world

    def add_border_to_space(self, size: float):
        # world border lines (start coords, end coords)
        # bottom, left, top, right
        self.border = (
            np.array([[0, 0], [0, 0], [0, size], [size, 0]], dtype=np.float32),
            np.array(
                [[size, 0], [0, size], [size, size], [size, size]],
                dtype=np.float32,
            ),
        )

        for w_start, w_end in zip(*self.border):
            wall = pymunk.Segment(
                self.space.static_body,
                (w_start[0], w_start[1]),
                (w_end[0], w_end[1]),
                self.border_thickness,
            )
            wall.friction = 1.0
            wall.collision_type = self.get_collision_id() + 1
            wall.color = self.WALL_COLOR
            self.space.add(wall)

    def clip_position(self, position: Vec2d) -> Vec2d:
        return Vec2d(
            *np.clip([position[0], position[1]], [0, 0], [self.size, self.size])
        )

    def check_border_collisions(
        self, ray_start_coords: np.ndarray, ray_end_coords: np.ndarray
    ) -> np.ndarray:
        return self.check_line_line_intersection(
            ray_start_coords, ray_end_coords, *self.border
        )[0]


class CircularContinuousWorld(AbstractContinuousWorld):
    """A 2D continuous world with a circular border."""

    def copy(self) -> "CircularContinuousWorld":
        world = CircularContinuousWorld(
            size=self.size,
            blocks=self.blocks,
            interior_walls=self.interior_walls,
            agent_radius=self.agent_radius,
            border_thickness=self.border_thickness,
        )
        for id, (body, shape) in self.entities.items():
            # make copies of each entity, and ensure the copies are linked correctly
            # and added to the new world and world space
            body = body.copy()
            shape = shape.copy()
            shape.body = body
            world.space.add(body, shape)
            world.entities[id] = (body, shape)
        return world

    def add_border_to_space(self, size: float):
        num_segments = 180
        radius = size / 2
        segment_angle_inc = 2 * math.pi / num_segments

        self.border = []
        for i in range(num_segments):
            prev_angle = i * segment_angle_inc
            angle = (i + 1) * segment_angle_inc
            x1 = radius * math.cos(prev_angle) + radius
            y1 = radius * math.sin(prev_angle) + radius
            x2 = radius * math.cos(angle) + radius
            y2 = radius * math.sin(angle) + radius
            wall = pymunk.Segment(
                self.space.static_body,
                (x1, y1),
                (x2, y2),
                self.border_thickness,
            )
            wall.friction = 0.0
            wall.color = self.WALL_COLOR
            wall.collision_type = self.get_collision_id() + 1
            self.border.append(wall)

        for wall in self.border:
            self.space.add(wall)

        self.size = size

    def check_border_collisions(
        self, ray_start_coords: np.ndarray, ray_end_coords: np.ndarray
    ) -> np.ndarray:
        center = np.array([0, 0]).reshape(1, 2)
        return self.check_circle_line_intersection(
            center, np.array([self.size]), ray_start_coords, ray_end_coords
        )

    def clip_position(self, position: Vec2d) -> Vec2d:
        x, y = position.x, position.y
        r_max = self.size / 2
        radius = np.linalg.norm([x - r_max, y - r_max])
        gamma = math.atan2(y, x)
        if radius > r_max:
            x = r_max * (1 + math.cos(gamma))
            y = r_max * (1 + math.sin(gamma))

        return Vec2d(x, y)


def generate_interior_walls(
    width: int, height: int, blocked_coords: Iterable[Coord]
) -> List[Line]:
    """Generate interior walls for rectangular world based on blocked coordinates."""
    # dx, dy
    # north, east, south, west
    directions = [(0, -1), (1, 0), (0, 1), (-1, 0)]
    line_offsets = [
        ((0, 0), (1, 0)),
        ((1, 0), (1, 1)),
        ((0, 1), (1, 1)),
        ((0, 0), (0, 1)),
    ]

    # get line for each block face adjacent to empty cell
    lines_map: Dict[Coord, Set[Coord]] = {}
    lines: Set[IntLine] = set()
    for x, y in blocked_coords:
        for (dx, dy), line_offset in zip(directions, line_offsets):
            if (x + dx, y + dy) in blocked_coords:
                # adjacent cell blocked
                continue

            s_offset, e_offset = line_offset
            l_start = (x + s_offset[0], y + s_offset[1])
            l_end = (x + e_offset[0], y + e_offset[1])
            if (
                (l_start[0] == 0 and l_end[0] == 0)
                or (l_start[0] == width and l_end[0] == width)
                or (l_start[1] == 0 and l_end[1] == 0)
                or (l_start[1] == height and l_end[1] == height)
            ):
                # line is along border
                continue

            if l_start not in lines_map:
                lines_map[l_start] = set()
            lines_map[l_start].add(l_end)
            lines.add((l_start, l_end))

    # merge lines
    merged_lines: List[IntLine] = []

    # lines l1 and l2 can merge if
    # 1. l1[1] == l2[0] and (l1[0][0] == l2[1][0] or l1[0][1] == l2[1][1]
    # or vice versa with l1 and l2 swapped
    # Assumes lines are parallel to either x or y axis

    # Basically a DFS
    stack = list(lines)
    stack.sort(reverse=True)

    visited: Set[IntLine] = set()
    while len(stack):
        line = stack.pop()
        if line in visited:
            continue

        visited.add(line)
        l_start, l_end = line
        line_extended = True
        while l_end in lines_map and line_extended:
            line_extended = False
            for l_end_next in lines_map[l_end]:
                if l_end_next != l_start and (
                    l_start[0] == l_end_next[0] or l_start[1] == l_end_next[1]
                ):
                    # merge
                    visited.add((l_end, l_end_next))
                    visited.add((l_end_next, l_end))
                    l_end = l_end_next
                    line_extended = True

        merged_lines.append((l_start, l_end))

    return [
        ((float(ln[0][0]), float(ln[0][1])), (float(ln[1][0]), float(ln[1][1])))
        for ln in merged_lines
    ]



"""Random number generator functions.

Ref:
https://github.com/Farama-Foundation/Gymnasium/blob/v0.27.0/gymnasium/utils/seeding.py

"""
import random
from typing import Optional, Tuple, Union

import numpy as np

from posggym import error


RNG = Union[random.Random, np.random.Generator]


def np_random(seed: Optional[int] = None) -> Tuple[np.random.Generator, int]:
    """Create a numpy random number generator.

    Arguments
    ---------
    seed : the seed used to create the generator.

    Returns
    -------
    rng : the random number generator
    seed : the seed used for the rng (will equal argument seed if one is provided.)

    Raises
    ------
    Error: if seed is not None or a non-negative integer.

    """
    if seed is not None and not (isinstance(seed, int) and seed >= 0):
        if isinstance(seed, int) is False:
            raise error.Error(
                f"Seed must be a python integer, actual type: {type(seed)}"
            )
        else:
            raise error.Error(
                f"Seed must be greater or equal to zero, actual value: {seed}"
            )

    seed_seq = np.random.SeedSequence(seed)
    np_seed = seed_seq.entropy
    # np_seed should always be an int if seed is an int | None
    assert isinstance(np_seed, int)
    rng = np.random.Generator(np.random.PCG64(seed_seq))
    return rng, np_seed


def std_random(seed: Optional[int] = None) -> Tuple[random.Random, int]:
    """Create random number generator using python built-in `random.Random`.

    Arguments
    ---------
    seed : the seed used to create the generator.

    Returns
    -------
    rng : the random number generator
    seed : the seed used for the rng (will equal argument seed if one is provided.)

    Raises
    ------
    Error: if seed is not None or a non-negative integer.

    """
    if seed is not None and not (isinstance(seed, int) and seed >= 0):
        if isinstance(seed, int) is False:
            raise error.Error(
                f"Seed must be a python integer, actual type: {type(seed)}"
            )
        else:
            raise error.Error(
                f"Seed must be greater or equal to zero, actual value: {seed}"
            )

    if seed is None:
        # We use the np.random library to generate the seed, since currently no well
        # documented/standard way to do this with the builtin random library
        seed_seq = np.random.SeedSequence(None)
        seed = seed_seq.entropy  # type: ignore
        # seed_seq.entropy should always be an int if seed is an int | None
        assert isinstance(seed, int)

    rng = random.Random(seed)
    return rng, seed



"""Root '__init__' of the posggym package."""
# isort: skip_file
from posggym.agents.registration import (
    make,
    pprint_registry,
    register,
    register_spec,
    registry,
    spec,
    get_all_envs,
    get_all_env_policies,
    get_env_agent_policies,
)
from posggym.agents.policy import Policy
from posggym.agents import random_policies
from posggym.agents.continuous import (
    drone_team_capture,
    driving_continuous,
    pursuit_evasion_continuous,
    predator_prey_continuous,
)
from posggym.agents.grid_world import (
    driving,
    level_based_foraging,
    predator_prey,
    pursuit_evasion,
)


__all__ = [
    # core classes
    "Policy",
    # registration
    "make",
    "pprint_registry",
    "register",
    "register_spec",
    "registry",
    "spec",
    "get_all_envs",
    "get_all_env_policies",
    "get_env_agent_policies",
]


# Generic Random Policies
# ------------------------------

register(
    policy_name="Random",
    entry_point=random_policies.RandomPolicy,
    version=0,
    env_id=None,
    env_args=None,
    valid_agent_ids=None,
    nondeterministic=False,
)

register(
    policy_name="DiscreteFixedDistributionPolicy",
    entry_point=random_policies.DiscreteFixedDistributionPolicy,
    version=0,
    env_id=None,
    env_args=None,
    valid_agent_ids=None,
    nondeterministic=False,
)

# Continuous
# ----------

# Drone Team Capture
for policy_spec in drone_team_capture.policy_specs.values():
    register_spec(policy_spec)

for policy_spec in driving_continuous.policy_specs.values():
    register_spec(policy_spec)

for policy_spec in predator_prey_continuous.policy_specs.values():
    register_spec(policy_spec)

for policy_spec in pursuit_evasion_continuous.policy_specs.values():
    register_spec(policy_spec)


# Grid World
# ----------

# Driving
for policy_spec in driving.policy_specs.values():
    register_spec(policy_spec)


# Level Based Foraging
for policy_spec in level_based_foraging.policy_specs.values():
    register_spec(policy_spec)


# Predator Prey
for policy_spec in predator_prey.policy_specs.values():
    register_spec(policy_spec)


# Pursuit Evasion
for policy_spec in pursuit_evasion.policy_specs.values():
    register_spec(policy_spec)



"""Heuiristic policies for DroneTeamCapture env."""
from __future__ import annotations

import abc
import math
from typing import cast

import numpy as np

from posggym import logger
from posggym.agents.policy import Policy, PolicyID, PolicyState
from posggym.agents.utils import action_distributions
from posggym.envs.continuous.drone_team_capture import (
    DroneTeamCaptureModel,
    DTCAction,
    DTCObs,
)


class DTCHeuristicPolicy(Policy[DTCAction, DTCObs], abc.ABC):
    """Heuristic Policies for the Drone Team Capture continuous environment.

    This is the abstract Drone Team Capture heuristic policy class. Concrete
    implementations must implement the _get_action method.

    """

    def __init__(
        self,
        model: DroneTeamCaptureModel,
        agent_id: str,
        policy_id: PolicyID,
    ):
        if model.n_com_pursuers < model.n_pursuers - 1 or (
            model.observation_limit is not None
            and model.observation_limit < 2 * model.r_arena
        ):
            logger.warn(
                "The DroneTeamCapture Heuristic policies are designed for the case "
                "where each pursuer can see every other pursuer "
                "(i.e. `n_com_pursuers = n_pursuers - 1` and `observation_limit = None`"
                "). Using it for other cases may result in unexpected behavior. "
                f"Currently `n_com_pursuers = {model.n_com_pursuers}`,"
                f"`n_pursuers = {model.n_pursuers}`, and "
                f"`observation_limit = {model.observation_limit}`."
            )
        self.obs_dim = model.obs_dim
        self.omega_max = model.dyaw_limit
        self.pursuer_vel = model.max_pursuer_vel
        self.pred_target_vel = model.max_pursuer_vel
        super().__init__(model, agent_id, policy_id)

    @abc.abstractmethod
    def _get_action(self, state: PolicyState) -> DTCAction:
        """Get the next action from the state.

        Must be implemented by subclasses.
        """
        raise NotImplementedError

    def get_initial_state(self) -> PolicyState:
        state = super().get_initial_state()
        n = (self.obs_dim - 8) // 2
        state["prev_yaw"] = 0.0
        state["yaw"] = 0.0
        state["prev_xy"] = np.zeros((2,), dtype=np.float32)
        state["xy"] = np.zeros((2,), dtype=np.float32)
        state["prev_target_xy"] = np.zeros((2,), dtype=np.float32)
        state["target_xy"] = np.zeros((2,), dtype=np.float32)
        state["prev_other_xy"] = np.zeros((n, 2), dtype=np.float32)
        state["other_xy"] = np.zeros((n, 2), dtype=np.float32)
        return state

    def get_next_state(
        self,
        obs: DTCObs,
        state: PolicyState,
    ) -> PolicyState:
        # Unnormalize obs
        min_obs = self.model.observation_spaces[self.agent_id].low
        # need to change this to be the actual min distance
        # instead of -1.0 used in obs space to handle out-of-limit agents
        min_obs[5] = 0.0
        for i in range(9, len(min_obs), 2):
            min_obs[i] = 0.0

        obs = self.model.world.convert_into_interval(
            obs,
            min_obs,
            self.model.observation_spaces[self.agent_id].high,
            self.model.raw_obs_range[0],
            self.model.raw_obs_range[1],
            clip=False,
        )
        xy = obs[2:4]
        yaw = obs[0]

        target_xy = xy + np.array(
            [
                obs[5] * np.cos(obs[4] + yaw),
                obs[5] * np.sin(obs[4] + yaw),
            ],
            dtype=np.float32,
        )

        other_xy = (
            xy
            + np.array(
                [
                    obs[9::2] * np.cos(obs[8::2] + yaw),
                    obs[9::2] * np.sin(obs[8::2] + yaw),
                ],
                dtype=np.float32,
            ).T
        )

        return {
            "action": None,
            "prev_yaw": state["yaw"],
            "yaw": yaw,
            "prev_xy": state["xy"],
            "xy": xy,
            "prev_target_xy": state["target_xy"],
            "target_xy": target_xy,
            "prev_other_xy": state["other_xy"],
            "other_xy": other_xy,
        }

    def sample_action(self, state: PolicyState) -> DTCAction:
        return self._get_action(state)

    def get_pi(self, state: PolicyState) -> action_distributions.ActionDistribution:
        action = self._get_action(state)
        return action_distributions.DeterministicActionDistribution(action)

    def euclidean_dist(self, coord1: np.ndarray, coord2: np.ndarray) -> float:
        return float(np.linalg.norm(coord1[:2] - coord2[:2]))

    def rot(self, los_angle: float) -> np.ndarray:
        R = np.array(
            [
                [math.cos(los_angle), math.sin(los_angle)],
                [-math.sin(los_angle), math.cos(los_angle)],
            ]
        )
        return R

    def pp(self, alpha: float) -> float:
        Kpp = 100
        omega = Kpp * alpha
        omega = np.clip(omega, -self.omega_max, self.omega_max)
        return omega

    def normalise(self, dx: float, dy: float) -> np.ndarray:
        d = math.sqrt(dx**2 + dy**2)
        d = np.clip(d, 0.000001, 10000000)
        return np.array([dx, dy]) / d

    def normalise_safe(self, arr: np.ndarray) -> np.ndarray:
        d = np.linalg.norm(arr, axis=1)
        d = np.clip(d, 0.000001, 10000000)
        return (arr.T / d).T

    def get_value(self, state: PolicyState) -> float:
        raise NotImplementedError(
            f"`get_value()` not implemented by {self.__class__.__name__} policy"
        )


class DTCJanosovHeuristicPolicy(DTCHeuristicPolicy):
    """Drone Team Capture Janosov Heuristic policy.

    Uses the heuristic from the following paper:

    DTCJanosovHeuristicPolicy is the policy from this work:
    Janosov, M., Virágh, C., Vásárhelyi, G., & Vicsek, T. (2017). Group chasing tactics:
    how to catch a faster prey. New Journal of Physics, 19(5), 053003.
    doi:10.1088/1367-2630/aa69e7

    This implementation is a non-holonomic adaptation of the original work following
    the protocol described in:

    C. de Souza, R. Newbury, A. Cosgun, P. Castillo, B. Vidolov and D. Kulić,
    "Decentralized Multi-Agent Pursuit Using Deep Reinforcement Learning,"
    in IEEE Robotics and Automation Letters, vol. 6, no. 3, pp. 4552-4559,
    July 2021, doi: 10.1109/LRA.2021.3068952.

    """

    def _get_action(self, state: PolicyState) -> DTCAction:
        chase = self.attraction2(
            state["xy"],
            state["prev_xy"],
            state["target_xy"],
            state["prev_target_xy"],
        )

        inter = self.alignment2(
            state["xy"],
            state["prev_xy"],
            state["other_xy"],
            state["prev_other_xy"],
        )

        # Sum up the contributions to the desired velocity
        desired_velocity = chase + inter

        # Rotate the desired velocity according to the agent's heading angle
        R = self.rot(state["yaw"])
        direction = R.dot(desired_velocity)

        # Calculate the angle between the desired direction and the current heading
        alpha = cast(float, np.arctan2(direction[1], direction[0]))

        # Calculate the angular velocity (omega) using the pp method
        omega = self.pp(alpha)

        return np.array([omega], dtype=np.float32)

    def attraction2(
        self,
        xy_i: np.ndarray,
        prev_xy_i: np.ndarray,
        target_xy: np.ndarray,
        prev_target_xy: np.ndarray,
    ) -> np.ndarray:
        # Friction term
        dist = self.euclidean_dist(target_xy, xy_i)
        vel_p = xy_i - prev_xy_i
        vel_t = target_xy - prev_target_xy
        visc = (vel_p - vel_t) / dist**2

        # Atraction term
        target_pred = self.prediction(xy_i, target_xy, vel_t)
        atrac = self.normalise(*(target_pred - xy_i))

        chase = atrac + 1.5 * visc
        chase = self.normalise(chase[0], chase[1])

        return chase * self.pursuer_vel

    def alignment2(
        self,
        xy_i: np.ndarray,
        prev_xy_i: np.ndarray,
        other_xy: np.ndarray,
        prev_other_xy: np.ndarray,
    ):
        rad_inter = 250
        C_inter = 0.5
        C_f = 0.5

        d_ij = xy_i - other_xy
        d = np.linalg.norm(d_ij, axis=1)
        d_ij = self.normalise_safe(d_ij)

        vel_i = xy_i - prev_xy_i
        vel_j = other_xy - prev_other_xy

        rep = ((d_ij.T * np.clip(d - rad_inter, -10000, 0)) / d).T
        fric = ((vel_i - vel_j).T / d**2).T

        inte = self.normalise(*(-rep + C_f * fric).sum(axis=0))

        return C_inter * inte * self.pursuer_vel

    def prediction(
        self, xy_i: np.ndarray, target_xy: np.ndarray, vel_t: np.ndarray
    ) -> np.ndarray:
        dist = self.euclidean_dist(target_xy, xy_i)

        tau = 20
        time_pred = dist / self.pred_target_vel
        time_pred = np.clip(time_pred, 0, tau)

        vel_t = self.normalise(vel_t[0], vel_t[1]) * self.pred_target_vel
        pos_pred = target_xy + vel_t * time_pred

        return pos_pred


class DTCAngelaniHeuristicPolicy(DTCHeuristicPolicy):
    """Drone Team Capture Angelani Heuristic policy.

    Uses the heuristic from the following paper:

    DTCAngelaniHeuristicPolicy is the policy from this work:
    Angelani, L. (09 2012). Collective Predation and Escape Strategies.
    Physical Review Letters, 109. doi:10.1103/PhysRevLett.109.118104

    This implementation is a non-holonomic adaptation of the original work following
    the protocol described in:

    C. de Souza, R. Newbury, A. Cosgun, P. Castillo, B. Vidolov and D. Kulić,
    "Decentralized Multi-Agent Pursuit Using Deep Reinforcement Learning,"
    in IEEE Robotics and Automation Letters, vol. 6, no. 3, pp. 4552-4559,
    July 2021, doi: 10.1109/LRA.2021.3068952.

    """

    def _get_action(self, state: PolicyState) -> DTCAction:
        rep = self.repulsion(state["xy"], state["other_xy"])

        align = self.alignment(
            np.concatenate((state["xy"].reshape(1, -1), state["other_xy"]), axis=0),
            np.concatenate(
                (state["prev_xy"].reshape(1, -1), state["prev_other_xy"]), axis=0
            ),
        )
        atrac = self.attraction(state["xy"], state["target_xy"])

        vel = rep + align + atrac

        direction = self.rot(state["yaw"]).dot(vel)
        alpha = math.atan2(direction[1], direction[0])
        omega_i = self.pp(alpha)
        return np.array([omega_i], dtype=np.float32)

    def repulsion(self, xy_i: np.ndarray, other_xy: np.ndarray):
        r = 300
        # Vectorized calculation of relative positions
        r_iT = other_xy - xy_i
        dist = np.linalg.norm(r_iT, axis=1)  # Compute the distance for all pursuers
        mask = dist < r
        repulsive_agents = r_iT[mask]

        dxy = self.rep_force(repulsive_agents, dist[mask])

        return self.normalise(*dxy)

    def rep_force(self, r_ij: np.ndarray, dist: np.ndarray) -> np.ndarray:
        sigma = 3
        u = -r_ij / dist[:, None]
        # Need to use float64 for the exponential to avoid overflow
        den = 1 + np.exp((dist - 20) / sigma, dtype=np.float64)[:, None]
        rep = u / den
        return rep.sum(axis=0, dtype=np.float32)

    def alignment(
        self, pursuers_xy: np.ndarray, prev_pursuers_xy: np.ndarray
    ) -> np.ndarray:
        pursuers_diff = pursuers_xy - prev_pursuers_xy
        return self.normalise(*np.sum(pursuers_diff, axis=0))

    def attraction(self, xy_i: np.ndarray, target_xy: np.ndarray) -> np.ndarray:
        r_iT = target_xy - xy_i
        return self.normalise(*r_iT)


class DTCDPPHeuristicPolicy(DTCHeuristicPolicy):
    """Drone Team Capture Deviated Pure Pursuit Heuristic policy.

    Uses the heuristic from the following paper:

    Souza, C., Castillo, P., & Vidolov, B. (2022). Local interaction and navigation
    guidance for hunters drones: a chase behavior approach with real-time tests.
    Robotica, 40(8), 2697–2715.

    """

    def _get_action(self, state: PolicyState) -> DTCAction:
        offset = math.pi / 8
        sense = 0.0

        sense = np.sum(self.delta(state["xy"], state["other_xy"], state["target_xy"]))

        alphaiT = self.engagement(state["xy"], state["yaw"], state["target_xy"])
        omega_i = 0.6 * (alphaiT - sense * offset)
        omega_i = np.clip(omega_i, -self.omega_max, self.omega_max)
        return np.array([omega_i], dtype=np.float32)

    def delta(
        self, xy_i: np.ndarray, other_xy: np.ndarray, target_xy: np.ndarray
    ) -> float:
        T_p = target_xy - xy_i
        los_angle = math.atan2(T_p[1], T_p[0])

        R = np.array(
            [
                [math.cos(los_angle), math.sin(los_angle)],
                [-math.sin(los_angle), math.cos(los_angle)],
            ]
        )
        i_j = other_xy - xy_i
        i_j = np.dot(i_j, R.T)

        # Avoid dividing by zero
        sense = np.where(np.abs(i_j[:, 1]) != 0, i_j[:, 1] / np.abs(i_j[:, 1]), 0)
        return np.sum(sense)

    def engagement(
        self, xy_i: np.ndarray, yaw_i: float, target_xy: np.ndarray
    ) -> float:
        # Rotation matrix of yaw
        R = np.array(
            [[math.cos(yaw_i), math.sin(yaw_i)], [-math.sin(yaw_i), math.cos(yaw_i)]]
        )

        T_p = R.dot(target_xy - xy_i)
        alpha = math.atan2(T_p[1], T_p[0])
        return alpha


class DTCGreedyHeuristicPolicy(DTCHeuristicPolicy):
    """Drone Team Capture Greedy Heuristic policy.

    Simple policy which selects the action which minimize the distance to the target
    at each step, ignoring the other pursuers.

    """

    def _get_action(self, state: PolicyState) -> DTCAction:
        target_xy = state["target_xy"]
        xy_i = state["xy"]
        yaw_i = state["yaw"]

        R = self.rot(yaw_i)

        T_p = R.dot(target_xy - xy_i)
        alpha = math.atan2(T_p[1], T_p[0])
        omega_i = self.pp(alpha)
        return np.array([omega_i], dtype=np.float32)



"""The Continuous Predator-Prey Environment."""
import math
from itertools import product
from typing import Dict, List, NamedTuple, Optional, Set, Tuple, Union, cast

import numpy as np
from gymnasium import spaces

import posggym.model as M
from posggym import logger
from posggym.core import DefaultEnv
from posggym.envs.continuous.core import (
    CircleEntity,
    PMBodyState,
    Position,
    SquareContinuousWorld,
    clip_actions,
)
from posggym.utils import seeding


class PPState(NamedTuple):
    """A state in the Continuous Predator-Prey Environment."""

    predator_states: np.ndarray
    prey_states: np.ndarray
    prey_caught: np.ndarray


PPObs = np.ndarray
PPAction = np.ndarray


class PredatorPreyContinuousEnv(DefaultEnv[PPState, PPObs, PPAction]):
    """The Continuous Predator-Prey Environment.

    A co-operative 2D continuous world problem involving multiple predator agents
    working together to catch prey agent/s in the environment.

    Possible Agents
    ---------------
    Varied number

    State Space
    -----------
    Each state consists of:

    1. 2D array containing the state of each predator
    2. 2D array containing the state of each prey
    3. 1D array containing whether each prey has been caught or not (0=no, 1=yes)

    The state of each predator and prey is a 1D array containing their:

    - `(x, y)` coordinate
    - angle/yaw (in radians)
    - velocity in x and y directions
    - angular velocity (in radians)

    Action Space
    ------------
    Each agent's actions is made up of two parts. The first action component specifies
    the angular velocity in `[-pi/10, pi/10]`, and the second component specifies the
    linear velocity in `[0, 1]`.

    Observation Space
    -----------------
    Each agent observes a local circle around themselves as a vector. This is achieved
    by a series of 'n_sensors' lines starting at the agent which extend for a distance
    of 'obs_dist'. For each line the agent observes the closest entity (wall, predator,
    prey) along the line. This table enumerates the observation space:

    |        Index: [start, end)        | Description                       | Values |
    | :-------------------------------: | --------------------------------: | :----: |
    |           0 - n_sensors           | Wall distance for each sensor     | [0, d] |
    |    n_sensors - (2 * n_sensors)    | Predator distance for each sensor | [0, d] |
    | (2 * n_sensors) - (3 * n_sensors) | Prey distance for each sensor     | [0, d] |


    Where `d = obs_dist`.

    If an entity is not observed (i.e. there is none along the sensor's line or it
    isn't the closest entity to the observing agent along the line), The distance will
    be `obs_dist`.

    The sensor reading ordering is relative to the agent's direction. I.e. the values
    for the first sensor at indices `0`, `n_sensors`, `2*n_sensors` correspond to the
    distance reading to a wall/obstacle, predator, and prey, respectively, in the
    direction the agent is facing.

    Rewards
    -------
    There are two modes of play:

    1. *Fully cooperative*: All predators share a reward and each agent receives
    a reward of `1.0 / num_prey` for each prey capture, independent of which
    predator agent/s were responsible for the capture.

    2. *Mixed cooperative*: Predators only receive a reward if they were part
    of the prey capture, receiving `1.0 / num_prey` per capture.

    In both modes prey can only been captured when at least `prey_strength`
    predators are in adjacent cells, where `1 <= prey_strength <= num_predators`.

    Dynamics
    --------
    Actions of the predator agents are deterministic and consist of moving based on
    their angle and linear velocity.

    Prey move according to the following rules (in order of priority):

    1. if a predator is within `prey_obs_dist` distance then moves away from closest
        predator
    2. if another prey is within `prey_obs_dist` distance then moves away from closest
        prey
    3. else move randomly

    Where `prey_obs_dist = 0.9 * obs_dist`, meaning predators are able to see slightly
    further than prey. If a prey is caught it is removed from the world (i.e. it's
    coords become `(-1, -1)`).

    Starting State
    --------------
    Predators start from random separate locations along the edge of the world
    (either in a corner, or half-way along a side), while prey start at random locations
    around the middle of the world.

    Episodes End
    ------------
    Episodes ends when all prey have been captured. By default a `max_episode_steps`
    limit of `100` steps is also set. This may need to be adjusted when using larger
    worlds (this can be done by manually specifying a value for `max_episode_steps` when
    creating the environment with `posggym.make`).

    Arguments
    ---------

    - `world` - the world layout to use. This can either be a string specifying one of
        the supported worlds, or a custom :class:`PPWorld` object (default = `"10x10"`).
    - `num_predators` - the number of predator (and thus controlled agents)
        (default = `2`).
    - `num_prey` - the number of prey (default = `3`)
    - `cooperative` - whether agents share all rewards or only get rewards for prey they
        are involved in capturing (default = 'True`)
    - `prey_strength` - how many predators are required to capture each prey, minimum is
        `1` and maximum is `min(4, num_predators)`. If `None` this is set to
        `min(4, num_predators)` (default = 'None`)
    - `obs_dist` - the local observation distance, specifying how far away in each
        direction each predator and prey agent observes (default = `4`).
    - `n_sensors` - the number of lines eminating from the agent. The agent will observe
        at `n` equidistance intervals over `[0, 2*pi]` (default = `16`).

    Available variants
    ------------------

    The PredatorPrey environment comes with a number of pre-built world layouts which
    can be passed as an argument to `posggym.make`, to create different worlds. All
    layouts support 2 to 8 agents.

    | World name         | World size |
    |-------------------|-----------|
    | `5x5`             | 5x5       |
    | `5x5Blocks`       | 5x5       |
    | `10x10`           | 10x10     |
    | `10x10Blocks`     | 10x10     |
    | `15x15`           | 15x15     |
    | `15x15Blocks`     | 15x15     |
    | `20x20`           | 20x20     |
    | `20x20Blocks`     | 20x20     |


    For example to use the Predator Prey environment with the `15x15Blocks` world, 4
    predators, 4 prey, and episode step limit of 100, and the default values for the
    other parameters (`cooperative`, `obs_dist`, `prey_strength`) you would use:

    ```python
    import posggym
    env = posggym.make(
        'PredatorPreyContinuous-v0',
        max_episode_steps=100,
        world="15x15Blocks",
        num_predators=4,
        num_prey=4
    )
    ```

    Version History
    ---------------
    - `v0`: Initial version

    Reference
    ---------
    - Ming Tan. 1993. Multi-Agent Reinforcement Learning: Independent vs. Cooperative
      Agents. In Proceedings of the Tenth International Conference on Machine Learning.
      330–337.
    - J. Z. Leibo, V. F. Zambaldi, M. Lanctot, J. Marecki, and T. Graepel. 2017.
      Multi-Agent Reinforcement Learning in Sequential Social Dilemmas. In AAMAS,
      Vol. 16. ACM, 464–473
    - Lowe, Ryan, Yi I. Wu, Aviv Tamar, Jean Harb, OpenAI Pieter Abbeel, and Igor
      Mordatch. 2017. “Multi-Agent Actor-Critic for Mixed Cooperative-Competitive
      Environments.” Advances in Neural Information Processing Systems 30.

    """

    metadata = {
        "render_modes": ["human", "rgb_array"],
        "render_fps": 15,
    }

    def __init__(
        self,
        world: Union[str, "PPWorld"] = "10x10",
        num_predators: int = 2,
        num_prey: int = 3,
        cooperative: bool = True,
        prey_strength: Optional[int] = None,
        obs_dist: float = 4,
        n_sensors: int = 16,
        render_mode: Optional[str] = None,
    ):
        super().__init__(
            PredatorPreyContinuousModel(
                world=world,
                num_predators=num_predators,
                num_prey=num_prey,
                cooperative=cooperative,
                prey_strength=prey_strength,
                obs_dist=obs_dist,
                n_sensors=n_sensors,
            ),
            render_mode=render_mode,
        )
        self.window_surface = None
        self.clock = None
        self.window_size = 600
        self.draw_options = None
        self.world = None

    def render(self):
        if self.render_mode is None:
            assert self.spec is not None
            logger.warn(
                "You are calling render method without specifying any render mode. "
                "You can specify the render_mode at initialization, "
                f'e.g. posggym.make("{self.spec.id}", render_mode="rgb_array")'
            )
            return
        return self._render_img()

    def _render_img(self):
        import pygame
        from pymunk import Transform, pygame_util

        model = cast(PredatorPreyContinuousModel, self.model)
        state = cast(PPState, self.state)
        scale_factor = self.window_size / model.world.size

        if self.window_surface is None:
            pygame.init()
            if self.render_mode == "human":
                pygame.display.init()
                pygame.display.set_caption(self.__class__.__name__)
                self.window_surface = pygame.display.set_mode(
                    (self.window_size, self.window_size)
                )
            else:
                self.window_surface = pygame.Surface(
                    (self.window_size, self.window_size)
                )
            # Turn off alpha since we don't use it.
            self.window_surface.set_alpha(None)

        if self.clock is None:
            self.clock = pygame.time.Clock()

        if self.draw_options is None:
            pygame_util.positive_y_is_up = False
            self.draw_options = pygame_util.DrawOptions(self.window_surface)
            self.draw_options.transform = Transform.scaling(scale_factor)
            # don't render collision lines
            self.draw_options.flags = (
                pygame_util.DrawOptions.DRAW_SHAPES
                | pygame_util.DrawOptions.DRAW_CONSTRAINTS
            )

        if self.world is None:
            # get copy of model world, so we can use it for rendering without
            # affecting the original
            self.world = model.world.copy()

        for i, p_state in enumerate(state.predator_states):
            self.world.set_entity_state(f"pred_{i}", p_state)

        for i, p_state in enumerate(state.prey_states):
            self.world.set_entity_state(f"prey_{i}", p_state)

        # Need to do this for space to update with changes
        self.world.space.step(0.0001)

        # reset screen
        self.window_surface.fill(pygame.Color("white"))

        # draw sensor lines
        n_sensors = model.n_sensors
        for i, obs_i in self._last_obs.items():
            p_state = state.predator_states[int(i)]
            x, y, agent_angle = p_state[:3]
            angle_inc = 2 * math.pi / n_sensors
            for k in range(n_sensors):
                dist = min(obs_i[k], obs_i[n_sensors + k], obs_i[2 * n_sensors + k])
                angle = angle_inc * k + agent_angle
                end_x = x + dist * math.cos(angle)
                end_y = y + dist * math.sin(angle)
                scaled_start = (int(x * scale_factor), int(y * scale_factor))
                scaled_end = int(end_x * scale_factor), (end_y * scale_factor)

                pygame.draw.line(
                    self.window_surface, pygame.Color("red"), scaled_start, scaled_end
                )

        self.world.space.debug_draw(self.draw_options)

        if self.render_mode == "human":
            pygame.event.pump()
            pygame.display.update()
            self.clock.tick(self.metadata["render_fps"])
            return None

        return np.transpose(
            np.array(pygame.surfarray.pixels3d(self.window_surface)), axes=(1, 0, 2)
        )

    def close(self) -> None:
        if self.window_surface is not None:
            import pygame

            pygame.display.quit()
            pygame.quit()


class PredatorPreyContinuousModel(M.POSGModel[PPState, PPObs, PPAction]):
    """Predator-Prey Continuous Problem Model.

    Parameters
    ----------
    world_size : float
        the size of the world (height and width)
    num_predators : int
        the number of predator (and thus controlled agents)
    num_prey : int
        the number of prey
    cooperative : bool
        whether environment rewards are fully shared (True) or only awarded to
        capturing predators (i.e. mixed) (False)
    prey_strenth : int
        the minimum number of predators needed to capture a prey
    obs_dists : float
        number of cells in each direction around the agent that the agent can
        observe
    n_sensors : int
        the number of sensor lines for each predator

    """

    R_MAX = 1.0
    # prey agent move distance per step
    PREY_STEP_VEL = 0.5

    PREDATOR_COLOR = (55, 155, 205, 255)  # Blueish
    PREY_COLOR = (110, 55, 155, 255)  # purpleish

    def __init__(
        self,
        world: Union[str, "PPWorld"],
        num_predators: int,
        num_prey: int,
        cooperative: bool,
        prey_strength: Optional[int],
        obs_dist: float,
        n_sensors: int,
    ):
        assert 1 < num_predators <= 8
        assert num_prey > 0
        assert obs_dist > 0

        if prey_strength is None:
            prey_strength = min(4, num_predators)
        assert 0 < prey_strength <= min(4, num_predators)

        if isinstance(world, str):
            assert world in SUPPORTED_WORLDS, (
                f"Unsupported world name '{world}'. World name must be one of: "
                f"{list(SUPPORTED_WORLDS)}."
            )
            world = SUPPORTED_WORLDS[world]()
        # Cannot be a string by this point.
        self.world = cast(PPWorld, world)

        assert len(self.world.predator_start_positions) >= num_predators, (
            f"World of size ({self.world.size, self.world.size}) cannot support "
            f"{num_predators} predators. Try with less predators, using a larger world,"
            " or using a different world layout."
        )
        assert len(self.world.prey_start_positions) >= num_prey, (
            f"World of size ({self.world.size, self.world.size}) cannot support "
            f"{num_prey} prey. Try with less prey, using a larger world, or using a "
            "different world layout."
        )

        self.num_predators = num_predators
        self.num_prey = num_prey
        self.obs_dist = obs_dist
        # prey see 10 % less than predators
        self.prey_obs_dist = 0.9 * self.obs_dist
        self.cooperative = cooperative
        self.prey_strength = prey_strength
        self.per_prey_reward = self.R_MAX / self.num_prey
        self.n_sensors = n_sensors
        # capture radius large enough so prey in corner can be captured by 3 predators
        self.prey_capture_dist = 2.75 * self.world.agent_radius
        self.possible_agents = tuple((str(x) for x in range(self.num_predators)))

        def _pos_space(n_agents: int):
            # x, y, angle, vx, vy, vangle
            # stacked n_agents time
            # shape = (n_agents, 6)
            size, angle = self.world.size, 2 * math.pi
            low = np.array([-1, -1, -angle, -1, -1, -angle], dtype=np.float32)
            high = np.array(
                [size, size, angle, 1.0, 1.0, angle],
                dtype=np.float32,
            )
            return spaces.Box(
                low=np.tile(low, (n_agents, 1)), high=np.tile(high, (n_agents, 1))
            )

        self.state_space = spaces.Tuple(
            (
                # state of each predator
                _pos_space(self.num_predators),
                # state of each prey
                _pos_space(self.num_prey),
                # prey caught/not
                spaces.MultiBinary(self.num_prey),
            )
        )

        # can turn up to 45 degrees per step
        self.dyaw_limit = math.pi / 4
        self.action_spaces = {
            i: spaces.Box(
                low=np.array([-self.dyaw_limit, 0.0], dtype=np.float32),
                high=np.array([self.dyaw_limit, 1.0], dtype=np.float32),
            )
            for i in self.possible_agents
        }

        self.obs_dim = self.n_sensors * 3
        self.observation_spaces = {
            i: spaces.Box(
                low=0.0, high=self.obs_dist, shape=(self.obs_dim,), dtype=np.float32
            )
            for i in self.possible_agents
        }

        # All predators are identical so env is symmetric
        self.is_symmetric = True

        # Add physical entities to the world
        for i in range(self.num_predators):
            self.world.add_entity(f"pred_{i}", None, color=self.PREDATOR_COLOR)

        for i in range(self.num_prey):
            self.world.add_entity(f"prey_{i}", None, color=self.PREY_COLOR)

    @property
    def reward_ranges(self) -> Dict[str, Tuple[float, float]]:
        return {i: (0.0, self.R_MAX) for i in self.possible_agents}

    @property
    def rng(self) -> seeding.RNG:
        if self._rng is None:
            self._rng, seed = seeding.std_random()
        return self._rng

    def get_agents(self, state: PPState) -> List[str]:
        return list(self.possible_agents)

    def sample_initial_state(self) -> PPState:
        predator_positions = [*self.world.predator_start_positions]
        self.rng.shuffle(predator_positions)
        predator_states = np.zeros(
            (self.num_predators, PMBodyState.num_features()), dtype=np.float32
        )
        for i in range(self.num_predators):
            predator_states[i][:3] = predator_positions[i]

        prey_positions = [*self.world.prey_start_positions]
        self.rng.shuffle(prey_positions)
        prey_states = np.zeros(
            (self.num_prey, PMBodyState.num_features()), dtype=np.float32
        )
        for i in range(self.num_prey):
            prey_states[i][:3] = prey_positions[i]

        return PPState(
            predator_states,
            prey_states,
            np.zeros(self.num_prey, dtype=np.int8),
        )

    def sample_initial_obs(self, state: PPState) -> Dict[str, PPObs]:
        return self.get_obs(state)

    def step(
        self, state: PPState, actions: Dict[str, PPAction]
    ) -> M.JointTimestep[PPState, PPObs]:
        clipped_actions = clip_actions(actions, self.action_spaces)

        next_state = self._get_next_state(state, clipped_actions)
        obs = self.get_obs(next_state)
        rewards = self._get_rewards(state, next_state)

        all_done = all(next_state.prey_caught)
        truncated = {i: False for i in self.possible_agents}
        terminated = {i: all_done for i in self.possible_agents}

        info: Dict[str, Dict] = {i: {} for i in self.possible_agents}
        if all_done:
            for i in self.possible_agents:
                info[i]["outcome"] = M.Outcome.WIN

        return M.JointTimestep(
            next_state, obs, rewards, terminated, truncated, all_done, info
        )

    def _get_next_state(self, state: PPState, actions: Dict[str, PPAction]) -> PPState:
        prey_move_angles = self._get_prey_move_angles(state)

        # apply prey actions
        for i in range(self.num_prey):
            self.world.set_entity_state(f"prey_{i}", state.prey_states[i])
            if state.prey_caught[i]:
                # do nothing
                continue
            self.world.update_entity_state(
                f"prey_{i}",
                angle=prey_move_angles[i],
                vel=self.world.linear_to_xy_velocity(
                    self.PREY_STEP_VEL, prey_move_angles[i]
                ),
            )

        # apply predator actions
        for i in range(self.num_predators):
            action = actions[str(i)]
            self.world.set_entity_state(f"pred_{i}", state.predator_states[i])
            angle = state.predator_states[i][2] + action[0]
            self.world.update_entity_state(
                f"pred_{i}",
                angle=angle,
                vel=self.world.linear_to_xy_velocity(action[1], angle),
            )

        # simulate
        self.world.simulate(1.0 / 10, 10)

        # extract next state
        next_pred_states = np.array(
            [
                self.world.get_entity_state(f"pred_{i}")
                for i in range(self.num_predators)
            ],
            dtype=np.float32,
        )

        next_prey_states = np.array(
            [self.world.get_entity_state(f"prey_{i}") for i in range(self.num_prey)],
            dtype=np.float32,
        )

        next_prey_caught = state.prey_caught.copy()
        for i in range(self.num_prey):
            if state.prey_caught[i]:
                next_prey_states[i] = [-1.0, -1.0, 0.0, 0.0, 0.0, 0.0]
                continue
            pred_dists = np.linalg.norm(
                next_prey_states[i][:2] - next_pred_states[:, :2], axis=1
            )
            if (
                np.where(pred_dists <= self.prey_capture_dist, 1, 0).sum()
                >= self.prey_strength
            ):
                next_prey_caught[i] = 1
                next_prey_states[i] = [-1.0, -1.0, 0.0, 0.0, 0.0, 0.0]

        return PPState(next_pred_states, next_prey_states, next_prey_caught)

    def _get_prey_move_angles(self, state: PPState) -> List[float]:
        prey_actions = []
        active_prey = self.num_prey - state.prey_caught.sum()
        for i in range(self.num_prey):
            if state.prey_caught[i]:
                # prey stays in same position
                prey_actions.append(0.0)
                continue

            prey_state = state.prey_states[i]
            # try move away from predators
            pred_states = state.predator_states
            pred_dists = np.linalg.norm(prey_state[:2] - pred_states[:, :2], axis=1)
            min_pred_dist = pred_dists.min()
            if min_pred_dist <= self.prey_obs_dist:
                # get any predators within obs distance
                pred_idx = self.rng.choice(np.where(pred_dists == min_pred_dist)[0])
                pred_state = state.predator_states[pred_idx]
                angle = math.atan2(
                    prey_state[1] - pred_state[1], prey_state[0] - pred_state[0]
                )
                prey_actions.append(angle)
                continue

            if active_prey == 1:
                # no other prey to move away from so just move randomly
                angle = self.rng.uniform(0, 2 * math.pi)
                prey_actions.append(angle)
                continue

            # try move away from prey
            prey_dists = [
                np.linalg.norm(prey_state[:2] - p[:2])
                for j, p in enumerate(state.prey_states)
                if not state.prey_caught[j] and j != i
            ]
            min_prey_dist = min(prey_dists)
            if min_prey_dist <= self.prey_obs_dist:
                other_prey_idx = self.rng.choice(
                    np.where(prey_dists == min_prey_dist)[0]
                )
                other_prey_state = state.prey_states[other_prey_idx]
                angle = math.atan2(
                    prey_state[1] - other_prey_state[1],
                    prey_state[0] - other_prey_state[0],
                )
                prey_actions.append(angle)
                continue

            # move in random direction
            angle = self.rng.uniform(0, 2 * math.pi)
            prey_actions.append(angle)

        return prey_actions

    def get_obs(self, state: PPState) -> Dict[str, PPObs]:
        return {i: self._get_local_obs(i, state) for i in self.possible_agents}

    def _get_local_obs(self, agent_id: str, state: PPState) -> np.ndarray:
        state_i = state.predator_states[int(agent_id)]
        pos_i = (state_i[0], state_i[1], state_i[2])

        prey_coords = state.prey_states[state.prey_caught == 0, :2]
        prey_obs, _ = self.world.check_collision_circular_rays(
            pos_i,
            self.obs_dist,
            self.n_sensors,
            prey_coords,
            include_blocks=False,
            check_walls=False,
            use_relative_angle=True,
        )

        mask = np.ones(len(state.predator_states), dtype=bool)
        mask[int(agent_id)] = False
        pred_coords = state.predator_states[mask, :2]
        pred_obs, _ = self.world.check_collision_circular_rays(
            pos_i,
            self.obs_dist,
            self.n_sensors,
            pred_coords,
            include_blocks=False,
            check_walls=False,
            use_relative_angle=True,
        )

        obstacle_obs, _ = self.world.check_collision_circular_rays(
            pos_i,
            self.obs_dist,
            self.n_sensors,
            other_agents=None,
            include_blocks=True,
            check_walls=True,
            use_relative_angle=True,
        )

        obs = np.full((self.obs_dim,), self.obs_dist, dtype=np.float32)
        sensor_readings = np.vstack([obstacle_obs, pred_obs, prey_obs])
        min_idx = np.argmin(sensor_readings, axis=0)
        min_val = np.take_along_axis(
            sensor_readings, np.expand_dims(min_idx, axis=0), axis=0
        ).flatten()
        idx = np.ravel_multi_index(
            (min_idx, np.arange(self.n_sensors)), dims=sensor_readings.shape
        )
        obs[idx] = np.minimum(min_val, obs[idx])

        return obs

    def _get_rewards(self, state: PPState, next_state: PPState) -> Dict[str, float]:
        new_caught_prey = []
        for i in range(self.num_prey):
            if not state.prey_caught[i] and next_state.prey_caught[i]:
                new_caught_prey.append(state.prey_states[i])

        if len(new_caught_prey) == 0:
            return {i: 0.0 for i in self.possible_agents}

        if self.cooperative:
            reward = len(new_caught_prey) * (self.per_prey_reward)
            return {i: reward for i in self.possible_agents}

        rewards = {i: 0.0 for i in self.possible_agents}
        pred_states = next_state.predator_states
        for prey_state in new_caught_prey:
            pred_dists = np.linalg.norm(prey_state[:2] - pred_states[:, :2], axis=1)
            involved_predators = np.where(pred_dists <= self.prey_capture_dist)[0]
            predator_reward = self.per_prey_reward / len(involved_predators)
            for i in involved_predators:
                rewards[str(i)] += predator_reward

        return rewards


class PPWorld(SquareContinuousWorld):
    """A continuous 2D world for the Predator-Prey Problem."""

    def __init__(
        self,
        size: int,
        blocks: Optional[List[CircleEntity]],
        predator_start_positions: Optional[List[Position]] = None,
        prey_start_positions: Optional[List[Position]] = None,
    ):
        assert size >= 3
        super().__init__(
            size=size,
            blocks=blocks,
            interior_walls=None,
            agent_radius=0.4,
            border_thickness=0.1,
            enable_agent_collisions=True,
        )

        if predator_start_positions is None:
            predator_start_positions = []
            for col, row in product([0, size // 2, size - 1], repeat=2):
                if col not in (0, size - 1) and row not in (0, size - 1):
                    continue
                x, y = col + self.agent_radius, row + self.agent_radius
                invalid_pos = False
                for block_pos, block_size in self.blocks:
                    dist = self.euclidean_dist((x, y, 0.0), block_pos)
                    if dist <= self.agent_radius + block_size:
                        invalid_pos = True
                        break
                if not invalid_pos:
                    predator_start_positions.append((x, y, 0.0))

        self.predator_start_positions = predator_start_positions

        if prey_start_positions is None:
            # prey can start anywhere at least distance 2 * self.agent size away from
            # any predator (i.e. an agent wide gap from any predator)
            # Not very efficient, but only needs to be run once at the start
            prey_start_positions = []
            for col, row in product(range(size), range(size)):
                if col in (0, size - 1) or row in (0, size - 1):
                    continue
                x, y = col + self.agent_radius, row + self.agent_radius
                invalid_pos = False
                for pred_pos in self.predator_start_positions:
                    dist = self.euclidean_dist((x, y), pred_pos)
                    if dist < 2 * self.agent_radius:
                        invalid_pos = True
                        break
                if invalid_pos:
                    continue
                for block_pos, block_size in self.blocks:
                    dist = self.euclidean_dist((x, y), block_pos)
                    if dist < self.agent_radius + block_size:
                        invalid_pos = True
                        break
                if not invalid_pos:
                    prey_start_positions.append((x, y, 0.0))

        self.prey_start_positions = prey_start_positions

    def copy(self) -> "PPWorld":
        world = PPWorld(
            size=int(self.size),
            blocks=self.blocks,
            predator_start_positions=self.predator_start_positions,
            prey_start_positions=self.prey_start_positions,
        )
        for id, (body, shape) in self.entities.items():
            # make copies of each entity, and ensure the copies are linked correctly
            # and added to the new world and world space
            body = body.copy()
            shape = shape.copy()
            shape.body = body
            world.space.add(body, shape)
            world.entities[id] = (body, shape)
        return world


def parse_world_str(world_str: str) -> PPWorld:
    """Parse a str representation of a world into a world object.

    Notes on world str representation:

    . = empty/unallocated cell
    # = a block
    P = starting location for predator agents [optional] (defaults to edges)
    p = starting location for prey agent [optional] (defaults to center)

    Examples (" " quotes and newline chars omitted):

    1. A 10x10 world with 4 groups of blocks and using the default predator
       and prey start locations.

    ..........
    ..........
    ..##..##..
    ..##..##..
    ..........
    ..........
    ..##..##..
    ..##..##..
    ..........
    ..........

    2. Same as above but with predator and prey start locations defined for
    up to 8 predators and 4 prey. (This would be the default layout for the
    scenario where there are between 2 and 4 prey, i.e. if prey and predator
    start locations were left unspecified as in example 1.)

    P....P...P
    ..........
    ..##..##..
    ..##..##..
    ....pp....
    P...pp...P
    ..##..##..
    ..##..##..
    ..........
    P....P...P

    """
    row_strs = world_str.splitlines()
    assert len(row_strs) > 1
    assert all(len(row) == len(row_strs[0]) for row in row_strs)
    assert len(row_strs[0]) > 1
    assert len(row_strs) == len(row_strs[0])

    size = len(row_strs)
    blocks: Set[CircleEntity] = set()
    predator_coords = set()
    prey_coords = set()
    for r, c in product(range(size), repeat=2):
        # This is offset to the center of the square
        pos = (c + 0.5, r + 0.5, 0)
        char = row_strs[r][c]

        if char == "#":
            # Radius is 0.5
            blocks.add((pos, 0.5))
        elif char == "P":
            predator_coords.add(pos)
        elif char == "p":
            prey_coords.add(pos)
        else:
            assert char == "."

    return PPWorld(
        size,
        blocks=list(blocks),
        predator_start_positions=None
        if len(predator_coords) == 0
        else list(predator_coords),
        prey_start_positions=None if len(prey_coords) == 0 else list(prey_coords),
    )


def get_default_world(size: int, include_blocks: bool) -> PPWorld:
    """Get function for generaing default world with given size.

    If `include_blocks=True` then world will contain blocks with the following layout:

    .....
    .#.#.
    .....
    .#.#.
    .....

    Where `#` are the blocks, which will be represented as a single circle.
    """
    r = float(size / 10)
    if include_blocks:
        blocks = [
            ((x + r, y + r, 0.0), r)
            for x, y in product([size / 5, 3 * size / 5], repeat=2)
        ]
    else:
        blocks = []
    return PPWorld(size=size, blocks=blocks)


def get_5x5_world() -> PPWorld:
    """Generate 5x5 world layou`t."""
    return get_default_world(5, include_blocks=False)


def get_5x5_blocks_world() -> PPWorld:
    """Generate 5x5 Blocks world layout."""
    return get_default_world(5, include_blocks=True)


def get_10x10_world() -> PPWorld:
    """Generate 10x10 world layou`t."""
    return get_default_world(10, include_blocks=False)


def get_10x10_blocks_world() -> PPWorld:
    """Generate 10x10 Blocks world layout."""
    return get_default_world(10, include_blocks=True)


def get_15x15_world() -> PPWorld:
    """Generate 15x15 world layou`t."""
    return get_default_world(15, include_blocks=False)


def get_15x15_blocks_world() -> PPWorld:
    """Generate 15x15 Blocks world layout."""
    return get_default_world(15, include_blocks=True)


def get_20x20_world() -> PPWorld:
    """Generate 20x20 world layout."""
    return get_default_world(20, include_blocks=False)


def get_20x20_blocks_world() -> PPWorld:
    """Generate 20x20 Blocks world layout."""
    return get_default_world(20, include_blocks=True)


#  world: world_make_fn
SUPPORTED_WORLDS = {
    "5x5": get_5x5_world,
    "5x5Blocks": get_5x5_blocks_world,
    "10x10": get_10x10_world,
    "10x10Blocks": get_10x10_blocks_world,
    "15x15": get_15x15_world,
    "15x15Blocks": get_15x15_blocks_world,
    "20x20": get_20x20_world,
    "20x20Blocks": get_20x20_blocks_world,
}



"""Functions and classes for rendering the environment during episodes."""
import abc
from typing import Dict, Iterable

import posggym
import posggym.model as M
from posggym.agents.policy import Policy


class Renderer(abc.ABC):
    """Abstract Renderer Base class."""

    @abc.abstractmethod
    def render_step(
        self,
        episode_t: int,
        env: posggym.Env,
        timestep: M.JointTimestep,
        action: Dict[str, M.ActType],
        policies: Dict[str, Policy],
        episode_end: bool,
    ) -> None:
        """Render a single environment step."""


class EpisodeRenderer(Renderer):
    """Episode Renderer.

    Calls the posggym.Env.render() function.
    """

    def __init__(self, render_frequency: int = 1):
        self._render_frequency = render_frequency
        self._episode_count = 0

    def render_step(
        self,
        episode_t: int,
        env: posggym.Env,
        timestep: M.JointTimestep,
        action: Dict[str, M.ActType],
        policies: Dict[str, Policy],
        episode_end: bool,
    ) -> None:
        if self._episode_count % self._render_frequency != 0:
            self._episode_count += int(episode_end)
            return

        self._episode_count += int(episode_end)
        env.render()


class PauseRenderer(Renderer):
    """Pauses for user input after each step.

    Note, renderers are rendered in order so if you want to pause after all the
    other renderers are done for the step make sure to put the PauseRenderer
    instance at the end of the list of renderers that are passed to the
    generate_renders function, or handle calling each renderer manually.
    """

    def render_step(
        self,
        episode_t: int,
        env: posggym.Env,
        timestep: M.JointTimestep,
        action: Dict[str, M.ActType],
        policies: Dict[str, Policy],
        episode_end: bool,
    ) -> None:
        input("Press ENTER to continue.")


def generate_renders(
    renderers: Iterable[Renderer],
    episode_t: int,
    env: posggym.Env,
    timestep: M.JointTimestep,
    action: Dict[str, M.ActType],
    policies: Dict[str, Policy],
    episode_end: bool,
) -> None:
    """Handle the generation of environment step renderings."""
    for renderer in renderers:
        renderer.render_step(episode_t, env, timestep, action, policies, episode_end)



"""Functions and classes for running episodes."""
import logging
import time
from typing import Any, Dict, Iterable, List, NamedTuple, Optional

import posggym
import posggym.agents.evaluation.render as render_lib
import posggym.agents.evaluation.stats as stats_lib
import posggym.agents.evaluation.writer as writer_lib
import posggym.model as M
from posggym.agents.policy import Policy


LINE_BREAK = "-" * 60
MAJOR_LINE_BREAK = "=" * 60


class EpisodeLoopStep(NamedTuple):
    """Output for a single episode step."""

    env: posggym.Env
    timestep: M.JointTimestep
    actions: Dict[str, Any]
    policies: Dict[str, Policy]
    done: bool


def run_episode_loop(
    env: posggym.Env,
    policies: Dict[str, Policy],
) -> Iterable[EpisodeLoopStep]:
    """Run policies in environment."""
    assert len(policies) == len(
        env.agents
    ), f"{len(policies)} policies supplied for env with {len(env.agents)} agents."

    observations, infos = env.reset()
    joint_timestep = M.JointTimestep(
        state=env.state,
        observations=observations,
        rewards={i: 0.0 for i in env.agents},
        terminations={i: False for i in env.agents},
        truncations={i: False for i in env.agents},
        all_done=False,
        infos=infos,
    )

    init_action = {i: None for i in env.agents}
    yield EpisodeLoopStep(env, joint_timestep, init_action, policies, False)

    all_done = False
    steps = 0
    while not all_done:
        actions = {}
        for i in env.agents:
            actions[i] = policies[i].step(observations[i])

        observations, rewards, terminations, truncations, all_done, infos = env.step(
            actions
        )
        joint_timestep = M.JointTimestep(
            state=env.state,
            observations=observations,
            rewards=rewards,
            terminations=terminations,
            truncations=truncations,
            all_done=all_done,
            infos=infos,
        )
        steps += 1

        yield EpisodeLoopStep(env, joint_timestep, actions, policies, all_done)


def run_episode(
    env: posggym.Env,
    policies: Dict[str, Policy],
    num_episodes: int,
    trackers: List[stats_lib.Tracker],
    renderers: List[render_lib.Renderer],
    time_limit: Optional[int] = None,
    logger: Optional[logging.Logger] = None,
    writer: Optional[writer_lib.Writer] = None,
) -> stats_lib.AgentStatisticsMap:
    """Run Episode simulations for given env and policies."""
    logger = logging.getLogger() if logger is None else logger
    writer = writer_lib.NullWriter() if writer is None else writer

    logger.info(
        "%s\nRunning %d episodes with Time Limit = %s s\n%s",
        MAJOR_LINE_BREAK,
        num_episodes,
        str(time_limit),
        MAJOR_LINE_BREAK,
    )

    episode_num = 0
    progress_display_freq = max(1, num_episodes // 10)
    time_limit_reached = False
    run_start_time = time.time()

    for tracker in trackers:
        tracker.reset()

    while episode_num < num_episodes and not time_limit_reached:
        logger.log(
            logging.INFO - 1,
            "%s\nEpisode %d Start\n%s",
            MAJOR_LINE_BREAK,
            episode_num,
            MAJOR_LINE_BREAK,
        )

        for tracker in trackers:
            tracker.reset_episode()

        for policy in policies.values():
            policy.reset()

        timestep_sequence = run_episode_loop(env, policies)
        for t, loop_step in enumerate(timestep_sequence):
            for tracker in trackers:
                tracker.step(t, *loop_step)
            render_lib.generate_renders(renderers, t, *loop_step)

        if len(trackers):
            episode_statistics = stats_lib.generate_episode_statistics(trackers)
            writer.write_episode(episode_statistics)
            logger.log(
                logging.INFO - 1,
                "%s\nEpisode %d Complete\n%s",
                LINE_BREAK,
                episode_num,
                writer_lib.format_as_table(episode_statistics),
            )
        else:
            logger.log(logging.INFO - 1, "\nEpisode %d Complete\n", episode_num)

        if (episode_num + 1) % progress_display_freq == 0:
            logger.info("Episode %d / %d complete", episode_num + 1, num_episodes)

        episode_num += 1

        if time_limit is not None and time.time() - run_start_time > time_limit:
            time_limit_reached = True
            logger.info(
                "%s\nTime limit of %d s reached after %d episodes",
                MAJOR_LINE_BREAK,
                time_limit,
                episode_num,
            )

    if len(trackers):
        statistics = stats_lib.generate_statistics(trackers)
        logger.info(
            "%s\nSimulations Complete\n%s\n%s",
            MAJOR_LINE_BREAK,
            writer_lib.format_as_table(statistics),
            MAJOR_LINE_BREAK,
        )
    else:
        statistics = {}
        logger.info(
            "%s\nSimulations Complete\n%s",
            MAJOR_LINE_BREAK,
            MAJOR_LINE_BREAK,
        )

    return statistics



"""Heuristic policies for Level-Based Foraging env.

Reference:
https://github.com/uoe-agents/lb-foraging/blob/master/lbforaging/agents/heuristic_agent.py

"""
from __future__ import annotations

import random
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, cast

from posggym.agents.policy import Policy, PolicyID, PolicyState
from posggym.agents.utils import action_distributions
from posggym.envs.grid_world.level_based_foraging import (
    LBFAction,
    LBFObs,
    LevelBasedForagingModel,
)

if TYPE_CHECKING:
    from posggym.envs.grid_world.core import Coord


class LBFHeuristicPolicy(Policy[LBFAction, LBFObs]):
    """Heuristic agent for the Level-Based Foraging env.

    This is the abstract Level-Based Foraging env heuristic policy class.
    Concrete implementations must implement the get_action_from_obs method.
    """

    def __init__(
        self, model: LevelBasedForagingModel, agent_id: str, policy_id: PolicyID
    ):
        super().__init__(model, agent_id, policy_id)
        assert model.observation_mode in ("vector", "tuple")
        self._rng = random.Random()
        self.field_width, self.field_height = model.field_size

    def reset(self, *, seed: int | None = None):
        super().reset(seed=seed)
        if seed is not None:
            self._rng.seed(seed)

    def get_initial_state(self) -> PolicyState:
        state = super().get_initial_state()
        state["last_obs"] = None
        return state

    def get_next_state(
        self,
        obs: LBFObs,
        state: PolicyState,
    ) -> PolicyState:
        return {"last_obs": obs}

    def sample_action(self, state: PolicyState) -> LBFAction:
        return self._get_action_from_obs(state["last_obs"])

    def get_pi(self, state: PolicyState) -> action_distributions.ActionDistribution:
        pi = self._get_pi_from_obs(state["last_obs"])
        return action_distributions.DiscreteActionDistribution(pi, self._rng)

    def get_value(self, state: PolicyState) -> float:
        raise NotImplementedError(
            f"`get_value()` no implemented by {self.__class__.__name__} policy"
        )

    def _get_action_from_obs(self, obs: LBFObs) -> LBFAction:
        """Get action from observation.

        Must be implemented by subclasses.
        """
        raise NotImplementedError

    def _get_pi_from_obs(self, obs: LBFObs) -> Dict[LBFAction, float]:
        """Get action distribution from observation.

        Must be implemented by subclasses.
        """
        raise NotImplementedError

    def _closest_food(
        self,
        agent_pos: Coord,
        food_obs: List[Tuple[int, int, int]],
        max_food_level: Optional[int] = None,
    ) -> Optional[Coord]:
        food_distances: Dict[int, List[Coord]] = {}
        for y, x, level in food_obs:
            if x == -1 or (max_food_level is not None and level > max_food_level):
                continue
            distance = (agent_pos[0] - y) ** 2 + (agent_pos[1] - x) ** 2
            if distance not in food_distances:
                food_distances[distance] = []
            food_distances[distance].append((y, x))

        if len(food_distances) == 0:
            # No food in sight
            return None

        return self._rng.choice(food_distances[min(food_distances)])

    def _center_of_agents(self, agent_pos: List[Coord]) -> Coord:
        y_mean = sum(coord[0] for coord in agent_pos) / len(agent_pos)
        x_mean = sum(coord[1] for coord in agent_pos) / len(agent_pos)
        return round(x_mean), round(y_mean)

    def _move_towards(
        self, agent_pos: Coord, target: Coord, allowed_actions: List[LBFAction]
    ) -> List[LBFAction]:
        y, x = agent_pos
        r, c = target

        valid_actions = []
        if r < y and LBFAction.NORTH in allowed_actions:
            valid_actions.append(LBFAction.NORTH)
        if r > y and LBFAction.SOUTH in allowed_actions:
            valid_actions.append(LBFAction.SOUTH)
        if c > x and LBFAction.EAST in allowed_actions:
            valid_actions.append(LBFAction.EAST)
        if c < x and LBFAction.WEST in allowed_actions:
            valid_actions.append(LBFAction.WEST)

        if valid_actions:
            return valid_actions
        else:
            raise ValueError("No simple path found")

    def _get_valid_move_actions(
        self, agent_obs: List[Tuple[int, int, int]]
    ) -> List[LBFAction]:
        y, x = agent_obs[0][:2]
        other_agent_pos = {o[:2] for o in agent_obs[1:] if o[0] > -1}

        valid_actions = []
        if y > 0 and (y - 1, x) not in other_agent_pos:
            valid_actions.append(LBFAction.NORTH)
        if y < self.field_height - 1 and (y + 1, x) not in other_agent_pos:
            valid_actions.append(LBFAction.SOUTH)
        if x < self.field_width - 1 and (y, x + 1) not in other_agent_pos:
            valid_actions.append(LBFAction.EAST)
        if x > 0 and (y, x - 1) not in other_agent_pos:
            valid_actions.append(LBFAction.WEST)

        return valid_actions

    def _get_actions_towards_food(
        self,
        agent_obs: List[Tuple[int, int, int]],
        center_pos: Coord,
        food_obs: List[Tuple[int, int, int]],
        max_food_level: Optional[int] = None,
    ) -> List[LBFAction]:
        closest_food = self._closest_food(center_pos, food_obs, max_food_level)
        if closest_food is None:
            actions = self._get_valid_move_actions(agent_obs)
            if actions:
                return actions
            return [LBFAction.NONE]

        food_y, food_x = closest_food
        y, x = agent_obs[0][:2]
        if (abs(food_y - y) + abs(food_x - x)) == 1:
            return [LBFAction.LOAD]

        valid_move_actions = self._get_valid_move_actions(agent_obs)
        try:
            return self._move_towards((y, x), (food_y, food_x), valid_move_actions)
        except ValueError:
            if valid_move_actions:
                return valid_move_actions
            return [LBFAction.NONE]

    def _parse_obs(
        self, obs: LBFObs
    ) -> Tuple[List[Tuple[int, int, int]], List[Tuple[int, int, int]]]:
        model = cast(LevelBasedForagingModel, self.model)
        return model.parse_obs(obs)


class LBFHeuristicPolicy1(LBFHeuristicPolicy):
    """Level-Based Foraging Heuristic Policy 1.

    This policy always goes to the closest observed food, irrespective of
    the foods level.
    """

    def _get_action_from_obs(self, obs: LBFObs) -> LBFAction:
        agent_obs, food_obs = self._parse_obs(obs)
        agent_pos = agent_obs[0][:2]
        actions = self._get_actions_towards_food(agent_obs, agent_pos, food_obs)
        return self._rng.choice(actions)

    def _get_pi_from_obs(self, obs: LBFObs) -> Dict[LBFAction, float]:
        agent_obs, food_obs = self._parse_obs(obs)
        agent_pos = agent_obs[0][:2]
        actions = self._get_actions_towards_food(agent_obs, agent_pos, food_obs)
        action_dist = {a: 0.0 for a in LBFAction}
        for a in actions:
            action_dist[a] = 1.0 / len(actions)
        return action_dist


class LBFHeuristicPolicy2(LBFHeuristicPolicy):
    """Level-Based Foraging Heuristic Policy 2.

    This policy goes towards the visible food that is closest to the centre of
    visible players, irrespective of food level.
    """

    def _get_action_from_obs(self, obs: LBFObs) -> LBFAction:
        agent_obs, food_obs = self._parse_obs(obs)
        other_agent_pos = [o[:2] for o in agent_obs[1:] if o[0] > -1]

        if not other_agent_pos:
            actions = self._get_valid_move_actions(agent_obs)
            if actions:
                return self._rng.choice(actions)
            return LBFAction.NONE

        center_pos = self._center_of_agents(other_agent_pos)
        actions = self._get_actions_towards_food(agent_obs, center_pos, food_obs)
        return self._rng.choice(actions)

    def _get_pi_from_obs(self, obs: LBFObs) -> Dict[LBFAction, float]:
        agent_obs, food_obs = self._parse_obs(obs)
        other_agent_pos = [o[:2] for o in agent_obs[1:] if o[0] > -1]

        if not other_agent_pos:
            # no visible agents
            actions = self._get_valid_move_actions(agent_obs)
            if not actions:
                actions = [LBFAction.NONE]
        else:
            center_pos = self._center_of_agents(other_agent_pos)
            actions = self._get_actions_towards_food(agent_obs, center_pos, food_obs)

        action_dist = {a: 0.0 for a in LBFAction}
        for a in actions:
            action_dist[a] = 1.0 / len(actions)
        return action_dist


class LBFHeuristicPolicy3(LBFHeuristicPolicy):
    """Level-Based Foraging Heuristic Policy 3.

    This policy goes towards the closest visible food with a compatible level.
    """

    def _get_action_from_obs(self, obs: LBFObs) -> LBFAction:
        agent_obs, food_obs = self._parse_obs(obs)
        agent_pos = agent_obs[0][:2]
        agent_level = agent_obs[0][2]
        actions = self._get_actions_towards_food(
            agent_obs, agent_pos, food_obs, agent_level
        )
        return self._rng.choice(actions)

    def _get_pi_from_obs(self, obs: LBFObs) -> Dict[LBFAction, float]:
        agent_obs, food_obs = self._parse_obs(obs)
        agent_pos = agent_obs[0][:2]
        agent_level = agent_obs[0][2]
        actions = self._get_actions_towards_food(
            agent_obs, agent_pos, food_obs, agent_level
        )
        action_dist = {a: 0.0 for a in LBFAction}
        for a in actions:
            action_dist[a] = 1.0 / len(actions)
        return action_dist


class LBFHeuristicPolicy4(LBFHeuristicPolicy):
    """Level-Based Foraging Heuristic Policy 4.

    This policy goes towards the visible food which is closest to all visible
    agents such that the sum of their and the policy agent's level is
    sufficient to load the food.
    """

    def _get_action_from_obs(self, obs: LBFObs) -> LBFAction:
        agent_obs, food_obs = self._parse_obs(obs)
        other_agent_pos = [o[:2] for o in agent_obs[1:] if o[0] > -1]

        if not other_agent_pos:
            actions = self._get_valid_move_actions(agent_obs)
            if actions:
                return self._rng.choice(actions)
            return LBFAction.NONE

        agent_level_sum = sum([o[2] for o in agent_obs if o[0] > -1])
        center_pos = self._center_of_agents(other_agent_pos)
        actions = self._get_actions_towards_food(
            agent_obs, center_pos, food_obs, agent_level_sum
        )
        return self._rng.choice(actions)

    def _get_pi_from_obs(self, obs: LBFObs) -> Dict[LBFAction, float]:
        agent_obs, food_obs = self._parse_obs(obs)
        other_agent_pos = [o[:2] for o in agent_obs[1:] if o[0] > -1]

        if not other_agent_pos:
            # no visible agents
            actions = self._get_valid_move_actions(agent_obs)
            if not actions:
                actions = [LBFAction.NONE]
        else:
            agent_level_sum = sum([o[2] for o in agent_obs if o[0] > -1])
            center_pos = self._center_of_agents(other_agent_pos)
            actions = self._get_actions_towards_food(
                agent_obs, center_pos, food_obs, agent_level_sum
            )

        action_dist = {a: 0.0 for a in LBFAction}
        for a in actions:
            action_dist[a] = 1.0 / len(actions)
        return action_dist



"""Functions and classes for registering and loading implemented environments.

Based on Farama Foundation Gymnasium, copied and adapted here to avoid issues with
gymnasiums global registry and difference in naming format.
https://github.com/Farama-Foundation/Gymnasium/blob/v0.27.0/gymnasium/envs/registration.py

"""
from __future__ import annotations

import contextlib
import copy
import difflib
import importlib
import re
import sys
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, Iterable, List, Tuple, TYPE_CHECKING

from posggym import error, logger
from posggym.wrappers import OrderEnforcing, PassiveEnvChecker, TimeLimit


if sys.version_info < (3, 10):
    import importlib_metadata as metadata  # type: ignore
else:
    import importlib.metadata as metadata


if TYPE_CHECKING:
    from posggym.core import Env


# [namespace/](env-name)-v(version)
ENV_ID_RE = re.compile(
    r"^(?:(?P<namespace>[\w:-]+)\/)?(?:(?P<name>[\w:.-]+?))(?:-v(?P<version>\d+))?$"
)


def load(name: str) -> Callable:
    """Loads environment with name and returns an environment creation function.

    Arguments
    ---------
    name: The environment name

    Returns
    -------
    entry_point: Environment creation function.

    """
    mod_name, attr_name = name.split(":")
    mod = importlib.import_module(mod_name)
    fn = getattr(mod, attr_name)
    return fn


def parse_env_id(env_id: str) -> Tuple[str | None, str, int | None]:
    """Parse environment ID string format.

    [namespace/](env-name)-v(version)    env-name is group 1, version is group 2

    Arguments
    ---------
    env_id: The environment id to parse

    Returns
    -------
    ns: The environment namespace
    name: The environment name
    version: The environment version

    Raises
    ------
    Error: If the environment id does not a valid environment regex

    """
    match = ENV_ID_RE.fullmatch(env_id)
    if not match:
        raise error.Error(
            f"Malformed environment ID: {env_id}. (Currently all IDs must be of the "
            "form [namespace/](env-name)-v(version) (namespace is optional))."
        )
    namespace, name, version = match.group("namespace", "name", "version")
    if version is not None:
        version = int(version)

    return namespace, name, version


def get_env_id(ns: str | None, name: str, version: int | None) -> str:
    """Get the full env ID given a name and (optional) version and namespace.

    Inverse of :meth:`parse_env_id`.

    Arguments
    ---------
    ns: The environment namespace
    name: The environment name
    version: The environment version

    Returns
    -------
    env_id: The environment id

    """
    full_name = name
    if version is not None:
        full_name += f"-v{version}"
    if ns is not None:
        full_name = ns + "/" + full_name
    return full_name


@dataclass
class EnvSpec:
    """A specification for creating environments with posggym.make.

    Attributes
    ----------
    id: The official environment ID.
    entry_point: Python entrypoint of the environment class (e.g. module.name:Class).
    reward_threshold: The reward threshold before the task is considered solved.
    nondeterministic: Whether this environment is non-deterministic even after seeding.
    max_episode_steps: The maximum number of steps that an episode can take before.
        truncation.
    order_enforce: Whether to wrap the environment in an orderEnforcing wrapper, that
        enforces the order of of `reset` before `step` and `render` functions.
    disable_env_checker: Whether to disable the environment checker wrapper in
        `posggym.make`, by default False (runs the environment checker)
    kwargs: Additional kwargs to pass to the environment class

    """

    id: str
    entry_point: Callable | str

    # Environment attributes
    reward_threshold: float | None = field(default=None)
    nondeterministic: bool = field(default=False)

    # Wrappers
    max_episode_steps: int | None = field(default=None)
    order_enforce: bool = field(default=True)
    disable_env_checker: bool = field(default=False)

    # Environment Arguments
    kwargs: Dict = field(default_factory=dict)

    # post-init attributes
    namespace: str | None = field(init=False)
    name: str = field(init=False)
    version: int | None = field(init=False)

    def __post_init__(self):
        """Extract the namespace, name and version from id.

        Is called after spec is created.
        """
        # Initialize namespace, name, version
        self.namespace, self.name, self.version = parse_env_id(self.id)

    def make(self, **kwargs: Any) -> Env:
        """Calls ``make`` using the environment spec and any keyword arguments."""
        # For compatibility purposes
        return make(self, **kwargs)


def _check_namespace_exists(ns: str | None):
    """Check if a namespace exists. If it doesn't, print a helpful error message."""
    if ns is None:
        return
    namespaces = {
        spec_.namespace for spec_ in registry.values() if spec_.namespace is not None
    }
    if ns in namespaces:
        return

    suggestion = (
        difflib.get_close_matches(ns, namespaces, n=1) if len(namespaces) > 0 else None
    )
    suggestion_msg = (
        f"Did you mean: `{suggestion[0]}`?"
        if suggestion
        else f"Have you installed the proper package for {ns}?"
    )

    raise error.NamespaceNotFound(f"Namespace {ns} not found. {suggestion_msg}")


def _check_name_exists(ns: str | None, name: str):
    """Check if env exists in namespace. If it doesn't, print helpful error message."""
    _check_namespace_exists(ns)
    names = {
        spec_.name.lower(): spec_.name
        for spec_ in registry.values()
        if spec_.namespace == ns
    }

    if name in names.values():
        return

    suggestion = difflib.get_close_matches(name.lower(), names, n=1)
    namespace_msg = f" in namespace {ns}" if ns else ""
    suggestion_msg = f"Did you mean: `{names[suggestion[0]]}`?" if suggestion else ""

    raise error.NameNotFound(
        f"Environment {name} doesn't exist{namespace_msg}. {suggestion_msg}"
    )


def _check_version_exists(ns: str | None, name: str, version: int | None):
    """Check if env version exists in namespace. Print helpful error message if not.

    This is a complete test whether an environment identifier is valid, and will
    provide the best available hints.

    Arguments
    ---------
    ns: The environment namespace
    name: The environment space
    version: The environment version

    Raises
    ------
    DeprecatedEnv: The environment doesn't exist but a default version does or the
        environment version is deprecated
    VersionNotFound: The ``version`` used doesn't exist

    """
    if get_env_id(ns, name, version) in registry:
        return

    _check_name_exists(ns, name)
    if version is None:
        return

    message = (
        f"Environment version `v{version}` for environment "
        f"`{get_env_id(ns, name, None)}` doesn't exist."
    )

    env_specs = [
        spec_
        for spec_ in registry.values()
        if spec_.namespace == ns and spec_.name == name
    ]
    env_specs = sorted(env_specs, key=lambda spec_: int(spec_.version or -1))

    default_spec = [spec_ for spec_ in env_specs if spec_.version is None]

    if default_spec:
        message += f" It provides the default version {default_spec[0].id}`."
        if len(env_specs) == 1:
            raise error.DeprecatedEnv(message)

    # Process possible versioned environments
    versioned_specs = [spec_ for spec_ in env_specs if spec_.version is not None]

    latest_spec = max(
        versioned_specs, key=lambda spec: spec.version, default=None  # type: ignore
    )
    if latest_spec is None or latest_spec.version is None:
        return

    if version > latest_spec.version:
        version_list_msg = ", ".join(f"`v{spec_.version}`" for spec_ in env_specs)
        message += f" It provides versioned environments: [ {version_list_msg} ]."
        raise error.VersionNotFound(message)

    if version < latest_spec.version:
        raise error.DeprecatedEnv(
            f"Environment version v{version} for `{get_env_id(ns, name, None)}` "
            f"is deprecated. Please use `{latest_spec.id}` instead."
        )


def find_highest_version(ns: str | None, name: str) -> int | None:
    """Finds the highest registered version of the environment in the registry."""
    version: list[int] = [
        spec_.version
        for spec_ in registry.values()
        if spec_.namespace == ns and spec_.name == name and spec_.version is not None
    ]
    return max(version, default=None)


# Global registry of environments. Meant to be accessed through `register` and `make`
registry: dict[str, EnvSpec] = {}
current_namespace: str | None = None


def _check_spec_register(spec: EnvSpec):
    """Checks whether spec is valid to be registered.

    Helper function for `register`.
    """
    global registry
    latest_versioned_spec = max(
        (
            spec_
            for spec_ in registry.values()
            if spec_.namespace == spec.namespace
            and spec_.name == spec.name
            and spec_.version is not None
        ),
        key=lambda spec_: int(spec_.version),  # type: ignore
        default=None,
    )

    unversioned_spec = next(
        (
            spec_
            for spec_ in registry.values()
            if spec_.namespace == spec.namespace
            and spec_.name == spec.name
            and spec_.version is None
        ),
        None,
    )

    if unversioned_spec is not None and spec.version is not None:
        raise error.RegistrationError(
            "Can't register the versioned environment "
            f"`{spec.id}` when the unversioned environment "
            f"`{unversioned_spec.id}` of the same name already exists."
        )
    elif latest_versioned_spec is not None and spec.version is None:
        raise error.RegistrationError(
            "Can't register the unversioned environment "
            f"`{spec.id}` when the versioned environment "
            f"`{latest_versioned_spec.id}` of the same name "
            f"already exists. Note: the default behavior is "
            f"that `posggym.make` with the unversioned environment "
            f"will return the latest versioned environment"
        )


def _check_metadata(metadata_: Dict):
    """Checks validity of metadata. Printing warnings if it's invalid."""
    if not isinstance(metadata_, dict):
        raise error.InvalidMetadata(
            f"Expect the environment metadata to be dict, actual type: {type(metadata)}"
        )

    render_modes = metadata_.get("render_modes")
    if render_modes is None:
        logger.warn(
            "The environment creator metadata doesn't include `render_modes`, "
            f"contains: {list(metadata_.keys())}"
        )
    elif not isinstance(render_modes, Iterable):
        logger.warn(
            "Expects the environment metadata render_modes to be a Iterable, actual "
            f"type: {type(render_modes)}"
        )


# Public API


@contextlib.contextmanager
def namespace(ns: str):
    """Context manager for modifying the current namespace."""
    global current_namespace
    old_namespace = current_namespace
    current_namespace = ns
    yield
    current_namespace = old_namespace


def register(
    id: str,
    entry_point: Callable | str,
    reward_threshold: float | None = None,
    nondeterministic: bool = False,
    max_episode_steps: int | None = None,
    order_enforce: bool = True,
    disable_env_checker: bool = False,
    **kwargs,
):
    """Register an environment with posggym.

    The `id` parameter corresponds to the name of the environment, with the syntax as
    follows:

        `(namespace)/(env_name)-v(version)` where `namespace` is optional.

    It takes arbitrary keyword arguments, which are passed to the `EnvSpec` constructor.

    Arguments
    ---------
    id: The environment id
    entry_point: The entry point for creating the environment
    reward_threshold: The reward threshold considered to have learnt an environment
    nondeterministic: If the environment is nondeterministic (even with knowledge of
        the initial seed and all actions)
    max_episode_steps: The maximum number of episodes steps before truncation. Used
        by the Time Limit wrapper.
    order_enforce: If to enable the order enforcer wrapper to ensure users run
        functions in the correct order
    disable_env_checker: Whether to disable the environment checker for the environment.
        Recommended to False.
    **kwargs: arbitrary keyword arguments which are passed to the environment
        constructor

    """
    global registry, current_namespace
    ns, name, version = parse_env_id(id)

    ns_id = ns
    if current_namespace is not None:
        if (
            kwargs.get("namespace") is not None
            and kwargs.get("namespace") != current_namespace
        ):
            logger.warn(
                f"Custom namespace `{kwargs.get('namespace')}` is being overridden by "
                f"namespace `{current_namespace}`. If you are developing a plugin you "
                "shouldn't specify a namespace in `register` calls. "
                "The namespace is specified through the entry point package metadata."
            )
        ns_id = current_namespace

    full_id = get_env_id(ns_id, name, version)

    new_spec = EnvSpec(
        id=full_id,
        entry_point=entry_point,
        reward_threshold=reward_threshold,
        nondeterministic=nondeterministic,
        max_episode_steps=max_episode_steps,
        order_enforce=order_enforce,
        disable_env_checker=disable_env_checker,
        **kwargs,
    )
    _check_spec_register(new_spec)
    if new_spec.id in registry:
        logger.warn(f"Overriding environment {new_spec.id} already in registry.")
    registry[new_spec.id] = new_spec


def make(
    id: str | EnvSpec,
    max_episode_steps: int | None = None,
    disable_env_checker: bool | None = None,
    **kwargs,
) -> Env:
    """Create an environment according to the given ID.

    To find all available environments use `posggym.envs.registry.keys()` for all valid
    ids.

    Arguments
    ---------
    id: Name of the environment. Optionally, a module to import can be included,
        eg. 'module:Env-v0'
    max_episode_steps: Maximum length of an episode (TimeLimit wrapper).
    disable_env_checker: Whether to run the env checker, None will default to the
            environment specification `disable_env_checker` (which is by default False,
            running the environment checker), otherwise will run according to this
            parameter (`True` = not run, `False` = run)
    kwargs: Additional arguments to pass to the environment constructor.

    Returns
    -------
    env: An instance of the environment.

    Raises
    ------
    Error: If the ``id`` doesn't exist then an error is raised

    """
    if isinstance(id, EnvSpec):
        spec_ = id
    else:
        module, id = (None, id) if ":" not in id else id.split(":")  # type: ignore
        if module is not None:
            try:
                importlib.import_module(module)
            except ModuleNotFoundError as e:
                raise ModuleNotFoundError(
                    f"{e}. Environment registration via importing a module failed. "
                    f"Check whether '{module}' contains env registration and can be "
                    "imported."
                ) from e

        spec_ = registry.get(id)  # type: ignore

        ns, name, version = parse_env_id(id)
        latest_version = find_highest_version(ns, name)
        if (
            version is not None
            and latest_version is not None
            and latest_version > version
        ):
            logger.warn(
                f"The environment {id} is out of date. You should consider "
                f"upgrading to version `v{latest_version}`."
            )

        if version is None and latest_version is not None:
            version = latest_version
            new_env_id = get_env_id(ns, name, version)
            spec_ = registry.get(new_env_id)  # type: ignore
            logger.warn(
                f"Using the latest versioned environment `{new_env_id}` "
                f"instead of the unversioned environment `{id}`."
            )

        if spec_ is None:
            _check_version_exists(ns, name, version)
            raise error.Error(f"No registered env with id: {id}")

    _kwargs = spec_.kwargs.copy()
    _kwargs.update(kwargs)

    if spec_.entry_point is None:
        raise error.Error(f"{spec_.id} registered but entry_point is not specified")
    elif callable(spec_.entry_point):
        env_creator = spec_.entry_point
    else:
        # Assume it's a string
        env_creator = load(spec_.entry_point)

    render_modes = None
    if hasattr(env_creator, "metadata"):
        _check_metadata(env_creator.metadata)
        render_modes = env_creator.metadata.get("render_modes")
    mode = _kwargs.get("render_mode")

    # TODO Johnny
    # Add support for attempting to apply applying HumanRendering/RenderCollection
    # wrappers (see gymnasium.envs.registration:make function)
    if mode is not None and render_modes is not None and mode not in render_modes:
        raise error.UnsupportedMode(
            f"The environment is being initialised with render_mode={mode} "
            f"that is not in the possible render_modes ({render_modes})."
        )

    try:
        env = env_creator(**_kwargs)
    except TypeError as e:
        raise e

    # Copies the environment creation specification and kwargs to add to the
    # environment's specification details
    spec_ = copy.deepcopy(spec_)
    spec_.kwargs = _kwargs
    env.unwrapped.spec = spec_
    env.unwrapped.model.spec = spec_

    # Run the environment checker as the lowest level wrapper
    if disable_env_checker is False or (
        disable_env_checker is None and spec_.disable_env_checker is False
    ):
        env = PassiveEnvChecker(env)

    # Add the order enforcing wrapper
    if spec_.order_enforce:
        env = OrderEnforcing(env)

    # Add the time limit wrapper
    if max_episode_steps is not None:
        env = TimeLimit(env, max_episode_steps)
    elif spec_.max_episode_steps is not None:
        env = TimeLimit(env, spec_.max_episode_steps)

    return env


def spec(env_id: str) -> EnvSpec:
    """Retrieve the spec for the given environment from the global registry.

    Arguments
    ---------
    env_id: the environment id.

    Returns
    -------
    spec: the environment spec from the global registry.

    Raises
    ------
    Error: if environment with given ``env_id`` doesn't exist in global registry.

    """
    spec_ = registry.get(env_id)
    if spec_ is None:
        ns, name, version = parse_env_id(env_id)
        _check_version_exists(ns, name, version)
        raise error.Error(f"No registered env with id: {env_id}")
    else:
        assert isinstance(spec_, EnvSpec)
        return spec_


def pprint_registry(
    _registry: dict = registry,
    num_cols: int = 3,
    exclude_namespaces: List[str] | None = None,
    disable_print: bool = False,
) -> str | None:
    """Pretty print the environments in the registry.

    Arguments
    ---------
    _registry: Environment registry to be printed.
    num_cols: Number of columns to arrange environments in, for display.
    exclude_namespaces: Exclude any namespaces from being printed.
    disable_print: Whether to return a string of all the namespaces and environment IDs
        instead of printing it to console.

    Returns
    -------
    return_str: formatted str representation of registry, if ``disable_print=True``,
        otherwise returns ``None``.

    """
    # Defaultdict to store environment names according to namespace.
    namespace_envs = defaultdict(lambda: [])
    max_justify = float("-inf")
    for env in _registry.values():
        namespace, _, _ = parse_env_id(env.id)
        if namespace is None:
            # Since namespace is currently none, use regex to obtain namespace from
            # entrypoints.
            env_entry_point = re.sub(r":\w+", "", env.entry_point)
            e_ep_split = env_entry_point.split(".")
            if len(e_ep_split) >= 3:
                # If namespace is of the format - posggym.envs.env_group.env_name:env_id
                # or posggym.envs.env_group:env_id
                idx = 2
                namespace = e_ep_split[idx]
            elif len(e_ep_split) > 1:
                # If namespace is of the format - posggym.env_group
                idx = 1
                namespace = e_ep_split[idx]
            else:
                # If namespace cannot be found, default to env id.
                namespace = env.id
        namespace_envs[namespace].append(env.id)
        max_justify = max(max_justify, len(env.id))

    # Iterate through each namespace and print environment alphabetically.
    return_str = ""
    for namespace, envs in namespace_envs.items():
        # Ignore namespaces to exclude.
        if exclude_namespaces is not None and namespace in exclude_namespaces:
            continue
        return_str += f"{'=' * 5} {namespace} {'=' * 5}\n"  # Print namespace.
        # Reference: https://stackoverflow.com/a/33464001
        for count, item in enumerate(sorted(envs), 1):
            return_str += (
                item.ljust(max_justify) + " "
            )  # Print column with justification.
            # Once all rows printed, switch to new column.
            if count % num_cols == 0 or count == len(envs):
                return_str = return_str.rstrip(" ") + "\n"
        return_str += "\n"

    if disable_print:
        return return_str

    print(return_str, end="")
    return None
